from matplotlib.pyplot import savefig
from lcapy import Circuit, sqrt, f, oo

Rs = 30
G = 1000

R1 = 100
R2 = (G - 1) * R1

Vn = 1e-9 * sqrt(3.5 / f + 1)
In = 1e-12 * sqrt(250 / f + 1)

T = 273 + 20
k_B = 1.38e-23

a = Circuit('opamp-noninverting-amplifier.sch')

an = a.noisy()

Vno = an[8].V.n(f)

Vno = Vno.limit('A', oo)

Vnov = Vno.subs({'R1': R1, 'R2': R2, 'In1': 0, 'In2': 0, 'Vn': Vn, 'Rs': Rs,
                 'k_B': k_B, 'T': 0})

Vnoi = Vno.subs({'R1': R1, 'R2': R2, 'In1': In, 'In2': In, 'Vn': 0, 'Rs': Rs,
                 'k_B': k_B, 'T': 0})

Vnor = Vno.subs({'R1': R1, 'R2': R2, 'In1': 0, 'In2': 0, 'Vn': 0, 'Rs': Rs,
                 'k_B': k_B, 'T': T})

Vnot = Vno.subs({'R1': R1, 'R2': R2, 'In1': In, 'In2': In, 'Vn': Vn, 'Rs': Rs,
                 'k_B': k_B, 'T': T})

flim = (0.1, 10e3)
ax = (Vnot * 1e9).plot(flim, loglog=True, label='total')
ax = (Vnov * 1e9).plot(flim, loglog=True, label='Vn', axes=ax)
ax = (Vnoi * 1e9).plot(flim, loglog=True, label='In', axes=ax)
ax = (Vnor * 1e9).plot(flim, loglog=True, label='R', axes=ax)
ax.set_ylabel('Noise voltage density nV/$\sqrt{\mathrm{Hz}}$')
ax.grid(True, 'both')
ax.legend()

savefig(__file__.replace('.py', '.png'), bbox_inches='tight')



"""
This module performs loop analysis.  It is primarily for showing
the equations rather than evaluating them.

Copyright 2019--2023 Michael Hayes, UCECE

"""

# TODO, handle current sources and dependent sources.

from .circuitgraph import CircuitGraph
from .expr import expr, equation, ExprTuple
from .systemequations import SystemEquations
import sympy as sym


class LoopAnalysis(object):
    """This performs for loop analysis.  Currently, it uses mesh analysis
    and so is only applicable to circuits with a planar topology.

    The API is likely to change since different invocations find
    different current loops.

    >>> from lcapy import Circuit, LoopAnalysis
    >>> cct = Circuit('''
    ... V1 1 0 {u(t)}; down
    ... R1 1 2; right=2
    ... L1 2 3; down=2
    ... W1 0 3; right
    ... W 1 5; up
    ... W 2 6; up
    ... C1 5 6; right=2
    ...''')

    To perform loop analysis in the time domain:

    >>> la = LoopAnalysis(cct)

    To display the equations found by applying KVL around each mesh:

    >>> la.mesh_equations().pprint()

    To display the system of equations (in matrix form) that needs to
    be solved:

    >>> la.matrix_equations().pprint()

    This only works for dc, ac, or laplace domains.  For example,

    >>> LoopAnalysis(cct.laplace()).matrix_equations().pprint()

    """

    @classmethod
    def from_circuit(cls, cct):

        return cls(cct)

    def __init__(self, cct):

        self.cct = cct
        self.cg = CircuitGraph.from_circuit(cct)

        source_groups = cct.independent_source_groups()

        if len(source_groups) > 1:
            # Could convert to time domain
            raise ValueError('Multiple independent source domains')
        elif len(source_groups) == 1:
            self.kind = list(source_groups)[0]
        else:
            self.kind = cct.kind

        self._equations = self._make_equations()

        self._unknowns = self.mesh_currents()
        self._y = matrix(self._unknowns)

    def loops(self):
        """Return list of loops.  Note, the loops can vary for different
        invocations of the LoopAnalysis class."""

        return self.cg.loops()

    def loops_by_cpt_name(self):
        """Return list of loops specified by cpt name."""

        return self.cg.loops_by_cpt_name()

    @property
    def num_loops(self):

        return len(self.cg.loops)

    def mesh_currents(self):

        if not self.cg.is_planar:
            raise ValueError('Circuit topology is not planar')

        loops = self.loops()
        Nloops = len(loops)

        mesh_currents = ExprList(
            [Iname('I_%d' % (m + 1), self.kind) for m in range(Nloops)])
        return mesh_currents

    def _add_mesh_currents(self, loop, loops, node_names, mesh_currents):

        current = 0

        # Find opposing currents in other meshes flowing through cpt.
        for n, loop2 in enumerate(loops):

            # Convert ['X', 'Y', 'Z'] to ['X', 'Y', 'Z', 'X'] to
            # avoid worrying about the wrap-around.
            loop2 = loop2.copy()
            loop2.append(loop2[0])

            if node_names[0] not in loop2 or node_names[1] not in loop2:
                continue

            # Determine polarity of cpt.
            for l in range(len(loop2) - 1):
                if (node_names[0] == loop2[l] and
                        node_names[1] == loop2[l + 1]):
                    current -= mesh_currents[n]
                    break
                elif (node_names[1] == loop2[l] and
                      node_names[0] == loop2[l + 1]):
                    current += mesh_currents[n]
                    break

        return current

    def _process_loop(self, loop, mesh_current, loops, mesh_currents):

        result = Vtype(self.kind)(0)

        # Convert ['A', 'B', 'C'] to ['A', 'B', 'C', 'A'] to
        # avoid worrying about the wrap-around.
        loop1 = loop.copy()
        loop1.append(loop1[0])
        for j in range(len(loop1) - 1):

            elt = self.cg.component(loop1[j], loop1[j + 1])
            # Skip wires.
            if elt is None:
                continue

            if elt.is_current_source:
                raise ValueError('TODO: handle current source in loop')
            elif elt.is_dependent_source:
                raise ValueError('Dependent sources not handled yet')

            # Map node names to equipotential node names.
            node_names = [self.cct.node_map[node_name]
                          for node_name in elt.node_names]

            if elt.is_voltage_source:
                v = elt.cpt.voltage_equation(mesh_current, self.kind)
            else:
                current = self._add_mesh_currents(loop, loops, node_names,
                                                  mesh_currents)
                v = -elt.cpt.voltage_equation(current, self.kind)

            is_reversed = node_names[0] == loop1[j] \
                and node_names[1] == loop1[j + 1]

            if is_reversed:
                v = -v

            result += v

        return result

    def _make_equations(self):

        if not self.cg.is_planar:
            raise ValueError('Circuit topology is not planar')

        loops = self.loops()
        Nloops = len(loops)

        mesh_currents = [Iname('I_%d' % (m + 1), self.kind)
                         for m in range(Nloops)]
        equations = {}

        for m, loop in enumerate(loops):

            result = self._process_loop(loop, mesh_currents[m],
                                        loops, mesh_currents)
            equations[mesh_currents[m]] = (result, expr(0))

        return equations

    def mesh_equations_list(self):
        """Return mesh equations as a list."""

        result = ExprList()

        for current, (lhs, rhs) in self._equations.items():
            result.append(equation(lhs, rhs))
        return result

    def mesh_equations(self):
        """Return mesh equations as a dict keyed by the mesh current."""

        result = ExprDict()

        for current, (lhs, rhs) in self._equations.items():
            result[current] = equation(lhs, rhs)
        return result

    def _analyse(self):

        if self.kind in ('t', 'time'):
            raise ValueError(
                'Cannot put time domain equations into matrix form.  '
                'Convert to dc, ac, or laplace domain first.')

        subsdict = {}
        for m, i in enumerate(self._unknowns):
            subsdict[i.expr] = 'X_X%d' % m

        exprs = []
        for node, (lhs, rhs) in self._equations.items():
            lhs = lhs.subs(subsdict).expr.expand()
            rhs = rhs.subs(subsdict).expr.expand()
            exprs.append(lhs - rhs)

        y = []
        for y1 in self._y:
            y.append(y1.subs(subsdict).expr)

        A, b = sym.linear_eq_to_matrix(exprs, *y)

        y = [y1.expr for y1 in self._y]
        return SystemEquations(A, b, y)

    @property
    def A(self):
        """Return A matrix where A y = b."""

        if not hasattr(self, '_sys'):
            self._sys = self._analyse()
        return matrix(self._sys.A)

    @property
    def b(self):
        """Return b vector where A y = b."""

        if not hasattr(self, '_sys'):
            self._sys = self._analyse()
        return matrix(self._sys.b)

    @property
    def y(self):
        """Return y vector where A y = b."""
        return self._y

    def matrix_equations(self, form='default', invert=False):
        """Return the equations in matrix form.

        Forms can be:
         'default'
         'A y = b'
         'b = A y'
         'Ainv b = y'
         'y = Ainv b'

        If `invert` is True, evaluate the matrix inverse."""

        if not hasattr(self, '_sys'):
            self._sys = self._analyse()
        return self._sys.format(form, invert)

    @property
    def unknowns(self):
        """Return tuple of the unknown voltages"""

        return ExprTuple(self.y)

    def solve_laplace(self):
        """Determine the unknown voltages using Laplace transforms and
        return as a dict"""

        from .sexpr import s

        unknowns = self.unknowns(s)
        return self.mesh_equations()(s).solve(unknowns)

    def pdb(self):
        """Enter the python debugger."""

        import pdb
        pdb.set_trace()
        return self


from .expr import ExprList, ExprDict, expr  # nopep8
from .current import Iname  # nopep8
from .voltage import Vtype  # nopep8
from .matrix import matrix  # nopep8



from matplotlib.pyplot import subplots, savefig
from numpy import arange
from numpy.random import randn
from lcapy import expr

e = expr('a * exp(-t  / tau) * u(t)')
tv = arange(1, 100)
vv = e.subs({'a': 1, 'tau': 10}).evaluate(tv) + randn(len(tv)) * 0.05
results = e.estimate(tv, vv, ranges={'a': (0, 10), 'tau': (1, 20)})

vv2 = results(tv)
fig, axes = subplots(1, figsize=(6, 3))
axes.plot(tv, vv, '.')
axes.plot(tv, vv2)
savefig(__file__.replace('.py', '.png'), bbox_inches='tight')



from lcapy import Circuit, t
from matplotlib.pyplot import savefig
from numpy import linspace


cct = Circuit()

cct.add('V1 1 0 step')
cct.add('L1 1 2 L1 0')
cct.add('C1 2 3')
cct.add('R1 3 0')

Vr = cct[3].V
vr = Vr(t)

vr.pprintans('v_R(t)')

vr = vr.subs({'L1':1, 'C1':0.1, 'R1':10, 'V1':1})

if True:
    tv = linspace(0, 2, 400)
    vr.plot(tv)
    
    savefig('circuit-VRLC2b-vr.png')



"""This module provides support for the inverse DFT.  It
calculates the bilateral inverse DFT transform using:

   s(n) = (1 / N) * \sum_{n=0}^{N-1} S(k) e^{j * 2 * \pi * n * k / N}

Copyright 2021 Michael Hayes, UCECE

"""

from .dft import DFTTransformer
from .matrix import Matrix
from .sym import j, pi
import sympy as sym


__all__ = ('IDFT', 'IDFTmatrix')


class InverseDFTTransformer(DFTTransformer):

    name = 'inverse DFT'
    is_inverse = True

    def noevaluate(self, expr, k, n):

        foo = expr * sym.exp(2 * j * pi * n * k / self.N)
        result = sym.Sum(foo, (n, 0, self.N - 1))
        result /= self.N
        return result

    def check(self, expr, k, n, N=None, **assumptions):

        if expr.has(n):
            self.error('Expression depends on n')
        super(InverseDFTTransformer, self).check(expr, k, n, N, **assumptions)


inverse_dft_transformer = InverseDFTTransformer()


def inverse_discrete_fourier_transform(expr, k, n, N, evaluate=True, **assumptions):
    """Compute bilateral inverse DFT transform of expr.

    Undefined functions such as X(k) are converted to x(n)
    """

    return inverse_dft_transformer.transform(expr, k, n, evaluate=evaluate,
                                             N=N, **assumptions)


def IDFT(expr, k, n, N, evaluate=True, **assumptions):
    """Compute bilateral inverse discrete Fourier transform of expr.

    Undefined functions such as X(k) are converted to x(n)
    """

    return inverse_dft_transformer.transform(expr, k, n, evaluate=evaluate,
                                             N=N, **assumptions)


def IDFTmatrix(N):
    """Return inverse DFT matrix of size `N` x `N`."""

    from .functions import exp

    w = exp(j * 2 * pi / N)

    a = Matrix.zeros(N)

    for row in range(N):
        for col in range(N):
            a[row, col] = w ** (row * col)
    return a / N



from lcapy import *
from lcapy.discretetime import *
from lcapy.latex import latex_format_label
import unittest


class LcapyTesterLatex(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_latex_subscripts(self):

        self.assertEqual(expr('V1').latex(), 'V_{1}', 'V1')
        self.assertEqual(expr('V_1').latex(), 'V_{1}', 'V_1')
        self.assertEqual(expr('V_C').latex(), 'V_{C}', 'V_C')

        self.assertEqual(expr('V_C1').latex(), 'V_{C_{1}}', 'V_C1')
        self.assertEqual(expr('V_C_1').latex(), 'V_{C_{1}}', 'V_C_1')

        self.assertEqual(expr('Vrms').latex(), 'V_{\\mathrm{rms}}', 'Vrms')
        self.assertEqual(expr('V_rms').latex(), 'V_{\\mathrm{rms}}', 'V_rms')

        self.assertEqual(expr('V_C1(t)').latex(), 'V_{C_{1}}(t)', 'V_C1(t)')
        self.assertEqual(expr('V_C_1(t)').latex(), 'V_{C_{1}}(t)', 'V_C_1(t)')

    def test_latex_greek(self):

        self.assertEqual(expr('alpha').latex(), '\\alpha', 'alpha')

        self.assertEqual(latex_format_label('alpha'), '$\\alpha$', 'alpha')
        self.assertEqual(latex_format_label(r'\alpha'), '$\\alpha$', r'\alpha')
        self.assertEqual(latex_format_label(r'\frac'), '$\\frac$', r'\frac')



def unique(alist):

    # Order preserving...  list(set(alist)) gives different results
    # for different runs.
    used = set()
    return [x for x in alist if x not in used and (used.add(x) or True)]


class Cnodes(dict):
    """Common nodes"""

    def __init__(self, nodes):

        super(Cnodes, self).__init__()
        for node in nodes:
            # Use tuple so hashable.
            self[node] = (node, )

    def link(self, n1, n2):
        """Make nodes n1 and n2 share common node"""

        set1 = self[n1]
        set2 = self[n2]

        newset = tuple(unique(set1 + set2))

        for n in self[n1]:
            self[n] = newset
        for n in self[n2]:
            self[n] = newset

    def all_nodes(self):

        return unique(self.values())



"""This module provides the SubNetlistSimplifyMixin class.

Copyright 2022--2023 Michael Hayes, UCECE

"""

from .expr import expr
from warnings import warn
from collections import OrderedDict


class NetlistSimplifyMixin:

    def _do_simplify_combine(self, string, subset, net,
                             explain=False, add=False, series=False):

        if explain:
            print(string % subset)

        subset_list = list(subset)

        if add:
            total = expr(0)
            for name in subset_list:
                total += expr(self.elements[name].cpt.args[0])
        else:
            total = expr(0)
            for name in subset_list:
                total += (1 / expr(self.elements[name].cpt.args[0]))
            total = 1 / total

        if explain:
            print('%s combined value = %s' % (subset, total))

        ic = None
        name = subset_list[0]
        elt = self.elements[name]
        if elt.cpt.has_ic:
            ic = expr(0)
            for name1 in subset_list:
                ic += expr(self.elements[name1].cpt.args[1])

            if explain:
                print('%s combined IC = %s' % (subset, ic))

        newname = self.namer(name[0] + 't', self.elements)
        net1 = elt._new_value(total, ic)
        parts = net1.split(' ', 1)
        net1 = newname + ' ' + parts[1]

        net.add(net1)
        net.remove(name)

        for name1 in subset_list[1:]:
            # Replace with wire or open-circuit.
            if series:
                net1 = self.elements[name1]._netmake_W()
            else:
                net1 = self.elements[name1]._netmake_O()

            # Avoid creating open-circuit components.
            if True and series:
                net.add(net1)

            net.remove(name1)

        return True

    def _check_ic(self, subset):

        subset = subset.copy()
        name = subset.pop()
        has_ic = self.elements[name].has_ic

        okay = True
        for name1 in subset:
            if self.elements[name1].has_ic != has_ic:
                warn('Incompatible initial conditions for %s and %s' %
                     (name, name1))
                okay = False
        if not has_ic:
            return okay
        ic = self.elements[name].cpt.args[1]
        for name1 in subset:
            if self.elements[name1].cpt.args[1] != ic:
                warn('Incompatible initial conditions for %s and %s' %
                     (name, name1))
                okay = False

        return okay

    def _simplify_combine_series(self, skip, explain=False):

        net = self.copy()
        changed = False

        for aset in net.in_series():
            aset -= skip
            subsets = net._find_combine_subsets(aset)
            for k, subset in subsets.items():
                if k == 'I':
                    warn('Netlist has current sources in series: %s' % subset)
                elif k in ('R', 'NR', 'L', 'V', 'Z'):
                    if k == 'L' and not self._check_ic(subset):
                        continue
                    changed |= self._do_simplify_combine('Can add in series: %s',
                                                         subset, net, explain, True, True)
                elif k in ('C', 'Y'):
                    changed |= self._do_simplify_combine('Can combine in series: %s',
                                                         subset, net, explain, False, True)
                else:
                    raise RuntimeError('Internal error')

        return net, changed

    def _simplify_combine_parallel(self, skip, explain=False):

        net = self.copy()
        changed = False

        for aset in net.in_parallel():
            aset -= skip
            subsets = net._find_combine_subsets(aset)
            for k, subset in subsets.items():
                if k == 'V':
                    warn('Netlist has voltage sources in parallel: %s' % subset)
                elif k in ('R', 'NR', 'L', 'Z'):
                    changed |= self._do_simplify_combine('Can combine in parallel: %s',
                                                         subset, net, explain, False, False)
                elif k in ('C', 'Y', 'I'):
                    if k == 'C' and not self._check_ic(subset):
                        continue
                    changed |= self._do_simplify_combine('Can add in parallel: %s',
                                                         subset, net, explain, True, False)
                else:
                    raise RuntimeError('Internal error')

        # TODO, remove dangling wires connected to the removed components.

        return net, changed

    def _simplify_redundant_series(self, skip, explain=False):

        net = self.copy()
        changed = False

        for aset in net.in_series():
            Iname = None
            for name in aset:
                cpt = self._elements[name]
                if cpt.type == 'I':
                    Iname = name
                    break
            if Iname is not None:
                for name in aset:
                    cpt = self._elements[name]
                    if cpt.type != 'I':
                        warn('Have redundant %s in series with %s' %
                             (name, Iname))

        return net, False

    def _simplify_redundant_parallel(self, skip, explain=False):

        net = self.copy()
        changed = False

        for aset in net.in_parallel():
            Vname = None
            for name in aset:
                cpt = self._elements[name]
                if cpt.type == 'V':
                    Vname = name
                    break
            if Vname is not None:
                for name in aset:
                    cpt = self._elements[name]
                    if cpt.type != 'V':
                        warn('Have redundant %s in parallel with %s' %
                             (name, Vname))

        return net, False

    def _keep_dangling(self, cpt, keep_nodes):

        for node in cpt.nodes:
            if node.is_dangling and node.name in keep_nodes:
                return True

        return False

    def _remove_dangling(self, skip, explain=False, keep_nodes=None):

        new = self._new()
        changed = False

        for cpt in self._elements.values():
            if (cpt.is_dangling and cpt.name not in skip
                    and not self._keep_dangling(cpt, keep_nodes)):
                if explain:
                    print('Removing dangling component %s' % cpt.name)
                changed = True
            else:
                new._add(cpt._copy())

        return new, changed

    def _remove_disconnected(self, skip, explain=False, keep_nodes=None):

        new = self._new()
        changed = False

        for cpt in self._elements.values():
            if (cpt.is_disconnected and cpt.name not in skip
                    and not self._keep_dangling(cpt, keep_nodes)):
                if explain:
                    print('Removing disconnected component %s' % cpt.name)
                changed = True
            else:
                new._add(cpt._copy())

        return new, changed

    def _simplify_series(self, skip, explain=False):

        net, changed = self._simplify_redundant_series(skip, explain)
        net, changed2 = net._simplify_combine_series(skip, explain)
        return net, changed or changed2

    def _simplify_parallel(self, skip, explain=False):

        net, changed = self._simplify_redundant_parallel(skip, explain)
        net, changed2 = net._simplify_combine_parallel(skip, explain)
        return net, changed or changed2

    def remove_dangling(self, select=None, ignore=None, passes=0, explain=False,
                        modify=True, keep_nodes=None):
        """Simplify a circuit by removing dangling components.

        This performs a number of passes specified by `passes`.  If zero,
        this iterates until no more simplifications can be performed.

        Note, if there are no circuits, e.g., a series network, then
        all the components will be removed.

        `select` is a list of component names to consider for simplification.
        If `None`, all components are considered.

        If `explain` is True, the reason for a simplification is printed.
        If `modify` is False, no modifications are performed.

        """

        return self.simplify(select=select, ignore=ignore, passes=passes,
                             explain=explain, modify=modify,
                             series=False, parallel=False, dangling=True,
                             keep_nodes=keep_nodes)

    def remove_dangling_wires(self, passes=0, explain=False,
                              modify=True, keep_nodes=None):
        """Simplify a circuit by removing dangling wires.

        This also removes disconnected open-circuit components
        after dangling wires have been removed.

        This performs a number of passes specified by `passes`.  If zero,
        this iterates until no more simplifications can be performed.

        If `explain` is True, the reason for a simplification is printed.
        If `modify` is False, no modifications are performed."""

        new = self.simplify(select=self.components.wires,
                            passes=passes, explain=explain,
                            modify=modify, series=False,
                            parallel=False, dangling=True, disconnected=False,
                            keep_nodes=keep_nodes)

        # Remove disconnected open-circuit components.
        return new.simplify(select=self.components.open_circuits,
                            passes=passes, explain=explain,
                            modify=modify, series=False,
                            parallel=False, dangling=False, disconnected=True,
                            keep_nodes=keep_nodes)

    def remove_disconnected(self, select=None, ignore=None, passes=0,
                            explain=False, modify=True, keep_nodes=None):
        """Simplify a circuit by removing disconnected components.

        This performs a number of passes specified by `passes`.  If zero,
        this iterates until no more simplifications can be performed.

        `select` is a list of component names to consider for simplification.
        If `None`, all components are considered.

        If `explain` is True, the reason for a simplification is printed.
        If `modify` is False, no modifications are performed."""

        return self.simplify(select=select, ignore=ignore, passes=passes,
                             explain=explain, modify=modify,
                             series=False, parallel=False, dangling=False,
                             disconnected=True, keep_nodes=keep_nodes)

    def simplify_series(self, select=None, ignore=None, passes=0,
                        explain=False, modify=True, keep_nodes=None):
        """Simplify a circuit by combining components in series.

        This performs a number of passes specified by `passes`.  If zero,
        this iterates until no more simplifications can be performed.

        `select` is a list of component names to consider for simplification.
        If `None`, all components are considered.

        If `None`, all components are considered.
        If `explain` is True, the reason for a simplification is printed.
        If `modify` is False, no modifications are performed."""

        return self.simplify(select=select, ignore=ignore, passes=passes,
                             explain=explain, modify=modify,
                             series=True, parallel=False, dangling=False,
                             keep_nodes=keep_nodes)

    def simplify_parallel(self, select=None, ignore=None, passes=0,
                          explain=False, modify=True, keep_nodes=None):
        """Simplify a circuit by combining components in parallel.

        This performs a number of passes specified by `passes`.  If zero,
        this iterates until no more simplifications can be performed.

        `select` is a list of component names to consider for simplification.
        If `None`, all components are considered.

        If `None`, all components are considered.
        If `explain` is True, the reason for a simplification is printed.
        If `modify` is False, no modifications are performed."""

        return self.simplify(select=select, ignore=ignore, passes=passes,
                             explain=explain, modify=modify,
                             series=False, parallel=True, dangling=False,
                             keep_nodes=keep_nodes)

    def simplify(self, select=None, ignore=None, passes=0, series=True,
                 parallel=True, dangling=False, disconnected=False,
                 explain=False, modify=True, keep_nodes=None):
        """Simplify a circuit by combining components in series and combining
        components in parallel.

        If `dangling` is True, then dangling components are removed.
        Dangling components are not in a circuit.

        If `disconnected` is True, then disconnected components are removed.
        Disconnected components are not connected to any other components.

        This performs a number of passes specified by `passes`.  If zero,
        this iterates until no more simplifications can be performed.

        `select` is a list of component names to consider for simplification.
        If `None`, all components are considered.

        If `explain` is True, the reason for a simplification is printed.
        If `modify` is False, no modifications are performed.

        See also `simplify_series`, `simplify_parallel`, `remove_dangling`,
        and `remove_dangling_wires`.

        """

        if keep_nodes is None:
            if '0' in self.nodes:
                keep_nodes = ['0']
            else:
                keep_nodes = []

        keep_nodes = [str(node) for node in keep_nodes]

        skip = set()

        if select is not None:
            skip = set(self._elements) - set(select)

        if ignore is not None:
            skip = skip.union(set(ignore))

        # Perhaps use num cpts?
        if passes == 0:
            passes = 100

        net = self
        for m in range(passes):
            changed = False
            if dangling:
                net, changed1 = net._remove_dangling(skip, explain, keep_nodes)
                changed = changed or changed1
            if disconnected:
                net, changed1 = net._remove_disconnected(
                    skip, explain, keep_nodes)
                changed = changed or changed1
            if series:
                net, changed1 = net._simplify_series(skip, explain)
                changed = changed or changed1
            if parallel:
                net, changed1 = net._simplify_parallel(skip, explain)
                changed = changed or changed1

            if not changed:
                break
        if not modify:
            return self

        return net



from matplotlib.pyplot import savefig
from lcapy import Circuit, j, f, s, degrees, pi

a = Circuit('opamp-voltage-follower-C-load.sch')
H = a.transfer(2, 0, 1, 0)

f1 = 10
f2 = 1e6
A0 = 1e6

A = A0 * (1 / (1 + j * f / f1)) * (1 / (1 + j * f / f2))
A = A.subs(f, s / (j * 2 * pi))

Ro = 40
C = 100e-9

H = H.subs({'A': A, 'Ro': Ro, 'C': C})

H.plot()

savefig(__file__.replace('.py', '.png'), bbox_inches='tight')



"""This module provides the NormAngularFourierDomainExpression class
to represent Omega-domain (normalized angular Fourier domain)
expressions.

Copyright 2021--2022 Michael Hayes, UCECE

"""

from __future__ import division
from .domains import NormAngularFourierDomain
from .inverse_fourier import inverse_fourier_transform
from .inverse_dtft import IDTFT
from .expr import Expr, expr, expr_make
from .fexpr import f
from .sym import ssym, tsym, fsym, pi, Omegasym
from .sym import nsym, dt
from .units import u as uu
from .utils import factor_const, remove_images
from sympy import Integral, Expr as symExpr


class NormAngularFourierDomainExpression(NormAngularFourierDomain, Expr):

    """Normalized angular Fourier domain expression or symbol."""

    var = Omegasym

    def __init__(self, val, **assumptions):

        check = assumptions.pop('check', True)
        assumptions['real'] = True
        super(NormAngularFourierDomainExpression,
              self).__init__(val, **assumptions)

        expr = self.expr
        if check and expr.has(ssym) and not expr.has(Integral):
            raise ValueError(
                'Omega-domain expression %s cannot depend on s' % expr)
        if check and expr.has(tsym) and not expr.has(Integral):
            raise ValueError(
                'Omega-domain expression %s cannot depend on t' % expr)

    def as_expr(self):
        return NormAngularFourierDomainExpression(self)

    def inverse_fourier(self, evaluate=True, **assumptions):
        """Attempt inverse Fourier transform."""

        expr = self.subs(2 * pi * fsym * dt)
        result = inverse_fourier_transform(
            expr.sympy, fsym, tsym, evaluate=evaluate)

        return self.change(result, 'time', units_scale=uu.Hz, **assumptions)

    def IFT(self, evaluate=True, **assumptions):
        """Convert to time domain.  This is an alias for inverse_fourier."""

        return self.inverse_fourier(evaluate=evaluate, **assumptions)

    def IDTFT(self, var=None, evaluate=True, **assumptions):
        """Convert to discrete-time domain using inverse discrete-time
        Fourier transform."""

        foo = self.subs(2 * pi * f * dt)
        result = IDTFT(foo.expr, self.var, nsym, evaluate=evaluate)

        return self.change(result, 'discrete time', units_scale=uu.Hz,
                           **assumptions)

    def time(self, **assumptions):
        return self.inverse_fourier(**assumptions)

    def fourier(self, **assumptions):
        """Convert to Fourier domain."""
        from .symbols import f

        result = self.subs(f / dt)
        return result

    def angular_fourier(self, **assumptions):
        """Convert to angular Fourier domain."""
        from .symbols import omega

        result = self.subs(omega / dt)
        return result

    def norm_fourier(self, **assumptions):
        """Convert to normalized Fourier domain."""
        from .symbols import F
        from .sym import dt

        result = self.subs(2 * pi * F / dt)
        return result

    def norm_angular_fourier(self, **assumptions):
        """Convert to normalized angular Fourier domain."""

        return self

    def laplace(self, **assumptions):
        """Determine one-side Laplace transform with 0- as the lower limit."""

        result = self.time(**assumptions).laplace()
        return result

    def phasor(self, **assumptions):
        """Convert to phasor domain."""

        return self.time(**assumptions).phasor(**assumptions)

    def plot(self, Wvector=None, plot_type=None, **kwargs):
        """Plot frequency response at values specified by `Wvector`.

        If `Wvector` is a tuple, this sets the frequency limits.

        `plot_type` - 'dB-phase', 'dB-phase-degrees', 'mag-phase',
        'mag-phase-degrees', 'real-imag', 'mag', 'phase',
        'phase-degrees', 'real', or 'imag'.

        The default `plot_type` for complex data is `dB-phase`.

        `kwargs` include:
        `axes` - the plot axes to use otherwise a new figure is created
        `xlabel` - the x-axis label
        `ylabel` - the y-axis label
        `ylabel2` - the second y-axis label if needed, say for mag and phase
        `xscale` - the x-axis scaling, say for plotting as ms
        `yscale` - the y-axis scaling, say for plotting mV
        `norm` - use normalized frequency
        `dbmin` - the smallest value to plot in dB (default -120)
        in addition to those supported by the matplotlib plot command.

        The plot axes are returned.  This is a tuple for magnitude/phase or
        real/imaginary plots.

        There are many plotting options, see lcapy.plot and
        matplotlib.pyplot.plot.

        For example:
            V.plot(Wvector, log_frequency=True)
            V.real.plot(Wvector, color='black')
            V.phase.plot(Wvector, color='black', linestyle='--')

        By default complex data is plotted as separate plots of
        magnitude (dB) and phase.

        """

        from .plot import plot_angular_frequency
        return plot_angular_frequency(self, Wvector, plot_type=plot_type,
                                      norm=True, **kwargs)

    def bode_plot(self, Wvector=None, unwrap=True, **kwargs):
        """Plot frequency response for a frequency-domain phasor as a Bode
        plot (but without the straight line approximations).  Wvector
        specifies the normalised angular frequencies.  If it is a
        tuple (f1, f2), it sets the frequency limits.  Since a
        logarithmic frequency scale is used, f1 must be greater than
        0.

        `unwrap` controls phase unwrapping (default True).

        For more info, see `plot`.

        """

        from .plot import plot_angular_bode
        return plot_angular_bode(self, Wvector, norm=True, unwrap=unwrap,
                                 **kwargs)

    def nyquist_plot(self, Wvector=None, log_frequency=False, **kwargs):
        """Plot frequency response as a Nyquist plot (imaginary part versus
        real part).  Wvector specifies the normalised angular
        frequencies.  If it is a tuple (f1, f2), it sets the frequency
        limits as (f1, f2).

        `npoints` set the number of plotted points.

        The unit circle is shown by default.  This can be disabled with `unitcircle=False`.

        """

        from .plot import plot_nyquist

        if Wvector is None:
            Wvector = (-pi, pi)
        return plot_nyquist(self, Wvector, log_frequency=log_frequency, **kwargs)

    def nichols_plot(self, Wvector=None, log_frequency=False, **kwargs):
        """Plot frequency response as a Nichols plot (dB versus phase).
        Wvector specifies the normalised angular frequencies.  If it
        is a tuple (f1, f2), it sets the frequency limits as (f1, f2).

        `npoints` set the number of plotted points.

        """

        from .plot import plot_nichols

        if Wvector is None:
            Wvector = (-pi, pi)
        return plot_nichols(self, Wvector, log_frequency=log_frequency, **kwargs)


def Omegaexpr(arg, **assumptions):
    """Create NormAngularFourierDomainExpression object.
    If `arg` is Omegasym return Omega"""

    if arg is Omegasym:
        return Omega
    return expr_make('norm angular fourier', arg, **assumptions)


Omega = NormAngularFourierDomainExpression('Omega')
Omega.units = uu.rad




# This file is automatically generated by makeclasses.py

from .units import units, u as uu
from .voltagemixin import VoltageMixin
from .currentmixin import CurrentMixin
from .admittancemixin import AdmittanceMixin
from .impedancemixin import ImpedanceMixin
from .transfermixin import TransferMixin
from .voltagesquaredmixin import VoltageSquaredMixin
from .currentsquaredmixin import CurrentSquaredMixin
from .admittancesquaredmixin import AdmittanceSquaredMixin
from .impedancesquaredmixin import ImpedanceSquaredMixin
from .powermixin import PowerMixin
from sympy import sqrt, S
from .kseq import DiscreteFourierDomainSequence
from .nseq import DiscreteTimeDomainSequence
from .zseq import ZDomainSequence

__all__ = ('DiscreteTimeDomainVoltageSequence', 'DiscreteTimeDomainCurrentSequence', 'DiscreteTimeDomainAdmittanceSequence', 'DiscreteTimeDomainImpedanceSequence', 'DiscreteTimeDomainTransferFunctionSequence', 'DiscreteTimeDomainVoltagesquaredSequence', 'DiscreteTimeDomainCurrentsquaredSequence', 'DiscreteTimeDomainAdmittancesquaredSequence', 'DiscreteTimeDomainImpedancesquaredSequence', 'DiscreteTimeDomainPowerSequence', 'DiscreteFourierDomainVoltageSequence', 'DiscreteFourierDomainCurrentSequence', 'DiscreteFourierDomainAdmittanceSequence', 'DiscreteFourierDomainImpedanceSequence', 'DiscreteFourierDomainTransferFunctionSequence', 'DiscreteFourierDomainVoltagesquaredSequence', 'DiscreteFourierDomainCurrentsquaredSequence', 'DiscreteFourierDomainAdmittancesquaredSequence', 'DiscreteFourierDomainImpedancesquaredSequence', 'DiscreteFourierDomainPowerSequence', 'ZDomainVoltageSequence', 'ZDomainCurrentSequence', 'ZDomainAdmittanceSequence', 'ZDomainImpedanceSequence', 'ZDomainTransferFunctionSequence', 'ZDomainVoltagesquaredSequence', 'ZDomainCurrentsquaredSequence', 'ZDomainAdmittancesquaredSequence', 'ZDomainImpedancesquaredSequence', 'ZDomainPowerSequence')



class DiscreteTimeDomainVoltageSequence(VoltageMixin, DiscreteTimeDomainSequence):
    """Discrete time-domain voltage (units V)."""

    _default_units = uu.volt


class DiscreteTimeDomainCurrentSequence(CurrentMixin, DiscreteTimeDomainSequence):
    """Discrete time-domain current (units A)."""

    _default_units = uu.ampere


class DiscreteTimeDomainAdmittanceSequence(AdmittanceMixin, DiscreteTimeDomainSequence):
    """Discrete time-domain admittance (units S)."""

    _default_units = uu.siemens


class DiscreteTimeDomainImpedanceSequence(ImpedanceMixin, DiscreteTimeDomainSequence):
    """Discrete time-domain impedance (units ohm)."""

    _default_units = uu.ohm


class DiscreteTimeDomainTransferFunctionSequence(TransferMixin, DiscreteTimeDomainSequence):
    """Discrete time-domain transfer (units )."""

    _default_units = S.One


class DiscreteTimeDomainVoltagesquaredSequence(VoltageSquaredMixin, DiscreteTimeDomainSequence):
    """Discrete time-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class DiscreteTimeDomainCurrentsquaredSequence(CurrentSquaredMixin, DiscreteTimeDomainSequence):
    """Discrete time-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class DiscreteTimeDomainAdmittancesquaredSequence(AdmittanceSquaredMixin, DiscreteTimeDomainSequence):
    """Discrete time-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class DiscreteTimeDomainImpedancesquaredSequence(ImpedanceSquaredMixin, DiscreteTimeDomainSequence):
    """Discrete time-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class DiscreteTimeDomainPowerSequence(PowerMixin, DiscreteTimeDomainSequence):
    """Discrete time-domain power (units W)."""

    _default_units = uu.watt


class DiscreteFourierDomainVoltageSequence(VoltageMixin, DiscreteFourierDomainSequence):
    """Discrete frequency-domain voltage (units V)."""

    _default_units = uu.volt


class DiscreteFourierDomainCurrentSequence(CurrentMixin, DiscreteFourierDomainSequence):
    """Discrete frequency-domain current (units A)."""

    _default_units = uu.ampere


class DiscreteFourierDomainAdmittanceSequence(AdmittanceMixin, DiscreteFourierDomainSequence):
    """Discrete frequency-domain admittance (units S)."""

    _default_units = uu.siemens


class DiscreteFourierDomainImpedanceSequence(ImpedanceMixin, DiscreteFourierDomainSequence):
    """Discrete frequency-domain impedance (units ohm)."""

    _default_units = uu.ohm


class DiscreteFourierDomainTransferFunctionSequence(TransferMixin, DiscreteFourierDomainSequence):
    """Discrete frequency-domain transfer (units )."""

    _default_units = S.One


class DiscreteFourierDomainVoltagesquaredSequence(VoltageSquaredMixin, DiscreteFourierDomainSequence):
    """Discrete frequency-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class DiscreteFourierDomainCurrentsquaredSequence(CurrentSquaredMixin, DiscreteFourierDomainSequence):
    """Discrete frequency-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class DiscreteFourierDomainAdmittancesquaredSequence(AdmittanceSquaredMixin, DiscreteFourierDomainSequence):
    """Discrete frequency-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class DiscreteFourierDomainImpedancesquaredSequence(ImpedanceSquaredMixin, DiscreteFourierDomainSequence):
    """Discrete frequency-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class DiscreteFourierDomainPowerSequence(PowerMixin, DiscreteFourierDomainSequence):
    """Discrete frequency-domain power (units W)."""

    _default_units = uu.watt


class ZDomainVoltageSequence(VoltageMixin, ZDomainSequence):
    """Z-domain voltage (units V)."""

    _default_units = uu.volt


class ZDomainCurrentSequence(CurrentMixin, ZDomainSequence):
    """Z-domain current (units A)."""

    _default_units = uu.ampere


class ZDomainAdmittanceSequence(AdmittanceMixin, ZDomainSequence):
    """Z-domain admittance (units S)."""

    _default_units = uu.siemens


class ZDomainImpedanceSequence(ImpedanceMixin, ZDomainSequence):
    """Z-domain impedance (units ohm)."""

    _default_units = uu.ohm


class ZDomainTransferFunctionSequence(TransferMixin, ZDomainSequence):
    """Z-domain transfer (units )."""

    _default_units = S.One


class ZDomainVoltagesquaredSequence(VoltageSquaredMixin, ZDomainSequence):
    """Z-domain voltagesquared (units V^2)."""

    _default_units = uu.volt**2


class ZDomainCurrentsquaredSequence(CurrentSquaredMixin, ZDomainSequence):
    """Z-domain currentsquared (units A^2)."""

    _default_units = uu.ampere**2


class ZDomainAdmittancesquaredSequence(AdmittanceSquaredMixin, ZDomainSequence):
    """Z-domain admittancesquared (units S^2)."""

    _default_units = uu.siemens**2


class ZDomainImpedancesquaredSequence(ImpedanceSquaredMixin, ZDomainSequence):
    """Z-domain impedancesquared (units ohm^2)."""

    _default_units = uu.ohm**2


class ZDomainPowerSequence(PowerMixin, ZDomainSequence):
    """Z-domain power (units W)."""

    _default_units = uu.watt

seqclasses = {'discrete time': {'undefined': DiscreteTimeDomainSequence,
'voltage': DiscreteTimeDomainVoltageSequence,
'current': DiscreteTimeDomainCurrentSequence,
'admittance': DiscreteTimeDomainAdmittanceSequence,
'impedance': DiscreteTimeDomainImpedanceSequence,
'transfer': DiscreteTimeDomainTransferFunctionSequence,
'voltagesquared': DiscreteTimeDomainVoltagesquaredSequence,
'currentsquared': DiscreteTimeDomainCurrentsquaredSequence,
'admittancesquared': DiscreteTimeDomainAdmittancesquaredSequence,
'impedancesquared': DiscreteTimeDomainImpedancesquaredSequence,
'power': DiscreteTimeDomainPowerSequence,
}
,'discrete fourier': {'undefined': DiscreteFourierDomainSequence,
'voltage': DiscreteFourierDomainVoltageSequence,
'current': DiscreteFourierDomainCurrentSequence,
'admittance': DiscreteFourierDomainAdmittanceSequence,
'impedance': DiscreteFourierDomainImpedanceSequence,
'transfer': DiscreteFourierDomainTransferFunctionSequence,
'voltagesquared': DiscreteFourierDomainVoltagesquaredSequence,
'currentsquared': DiscreteFourierDomainCurrentsquaredSequence,
'admittancesquared': DiscreteFourierDomainAdmittancesquaredSequence,
'impedancesquared': DiscreteFourierDomainImpedancesquaredSequence,
'power': DiscreteFourierDomainPowerSequence,
}
,'Z': {'undefined': ZDomainSequence,
'voltage': ZDomainVoltageSequence,
'current': ZDomainCurrentSequence,
'admittance': ZDomainAdmittanceSequence,
'impedance': ZDomainImpedanceSequence,
'transfer': ZDomainTransferFunctionSequence,
'voltagesquared': ZDomainVoltagesquaredSequence,
'currentsquared': ZDomainCurrentsquaredSequence,
'admittancesquared': ZDomainAdmittancesquaredSequence,
'impedancesquared': ZDomainImpedancesquaredSequence,
'power': ZDomainPowerSequence,
}
,}




"""This module provides classes for schematic layout.

Copyright 2014--2021 Michael Hayes, UCECE

"""
from .config import colours
from .cnodes import Cnodes

# The component placement algorithm is similar to a critical path
# method (CPM).  For most netlists it does a good job.  However, there
# are some loop constructs where it fails.

# An alternative approach is to form a system of linear equations
# where the unknowns are the component positions and the component
# stretch.  Most of the component stretches will be free variables and
# are set to zero stretch.  However, this algorithm will also fail for
# the same loop constructs as the graph algorithm,

# Here's a problem example:
#
# R1 1 2; right=2
# R2 2 3; right=1
# W 2 5; down=1
# W 4 5; right=0.5
# W 5 6; right=0.5
# C 4 7; down
# R 6 9; down
# W 7 8; right=0.5
# W 8 9; right=0.5
# W 8 11; down=1
# W 10 11; right
# W 11 12; right=2

# For the horizontal graph, the critical path is through nodes 1 and
# 12 but this not found due to the loop for nodes 2, 4, 6, and 8.


class Gedge(object):
    """Edge between common nodes"""

    def __init__(self, cpt, from_gnode, to_gnode, size, stretch=False):

        self.cpt = cpt
        self.from_gnode = from_gnode
        self.to_gnode = to_gnode
        self.size = size
        self.stretch = stretch
        self.stretched = False

    def __repr__(self):

        return '%s --%s%s--> %s' % (
            self.from_gnode.fmt_name, self.size, '*' if self.stretch else '',
            self.to_gnode.fmt_name)

    @property
    def name(self):
        if self.cpt is None:
            return 'dummy'
        return self.cpt.name


class Gnode(object):
    """Drawing node"""

    def __init__(self, name):

        self.name = name
        self.dist = 0
        self._pos = None
        self.fedges = []
        self.redges = []

    @property
    def pos(self):
        return self._pos

    @pos.setter
    def pos(self, value):

        if value is not None and self._pos is not None:
            raise RuntimeError('Changing node %s pos' % self)
        if False and value is not None:
            print('Setting node %s to pos %f' % (self, value))

        self._pos = value

    def add_fedge(self, edge):

        for edge1 in self.fedges:
            if (edge1.cpt == edge.cpt and
                edge1.to_gnode == edge.to_gnode and
                    edge1.from_gnode == edge.from_gnode):
                return

        self.fedges.append(edge)

    def add_redge(self, edge):

        for edge1 in self.redges:
            if (edge1.cpt == edge.cpt and
                edge1.to_gnode == edge.to_gnode and
                    edge1.from_gnode == edge.from_gnode):
                return

        self.redges.append(edge)

    @property
    def fmt_name(self):

        if isinstance(self.name, tuple):
            return '(' + ', '.join(self.name) + ')'
        return self.name

    def __repr__(self):

        return self.fmt_name


class GraphPath(list):
    """This is a list of edges defining a path through a graph."""

    @property
    def dist(self):

        distance = 0
        for edge in self:
            distance += edge.size
        return distance

    @property
    def stretches(self):

        stretches = 0
        for edge in self:
            if edge.stretch:
                stretches += 1
        return stretches

    def on(self, match_node):

        if self == []:
            return False

        for edge in self:
            if edge.from_gnode == match_node:
                return True

        return edge.to_gnode == match_node

    def to_dist(self, match_node):

        if match_node == self.from_gnode:
            return 0

        distance = 0
        for edge in self:
            distance += edge.size
            if edge.to_gnode == match_node:
                return distance
        raise ValueError('Node not on path')

    def from_dist(self, match_node):

        if match_node == self.to_gnode:
            return 0

        distance = 0
        for edge in reversed(self):
            distance += edge.size
            if edge.from_gnode == match_node:
                return distance
        raise ValueError('Node not on path')

    @property
    def gnodes(self):
        """Return list of nodes along path including end nodes."""

        nodes = []
        nodes.append(self.from_gnode)
        for edge in self:
            nodes.append(edge.to_gnode)
        return nodes

    @property
    def to_gnode(self):

        return self[-1].to_gnode

    @property
    def from_gnode(self):

        return self[0].from_gnode


class Graph(dict):
    """Drawing graph"""

    def __init__(self, name, nodes, debug=False):

        self.cnodes = Cnodes(nodes)
        self.name = name
        self.debug = debug

    def __repr__(self):

        s = ''
        for gnode in self.values():
            if gnode.fedges == []:
                s += '@%s %s\n' % (gnode.dist, gnode.fmt_name)
            else:
                for edge in gnode.fedges:
                    s += '@%s %s\n' % (gnode.dist, edge)
        return s

    def __getitem__(self, key):

        # If key is an integer, convert to a string.
        if isinstance(key, int):
            key = '%d' % key

        # Allow indexing by a node name rather than by cnode tuple.
        if key in self.cnodes:
            key = self.cnodes[key]

        return super(Graph, self).__getitem__(key)

    def link(self, n1, n2):
        """Make nodes n1 and n2 share common gnode"""

        self.cnodes.link(n1, n2)

    def add(self, cpt, n1, n2, size, stretch):
        """Add cpt between nodes n1 and n2 to the graph"""

        if size == 0:
            return

        if size < 0:
            n1, n2 = n2, n1
            size = -size

        if n1 in self.cnodes:
            n1 = self.cnodes[n1]
        if n2 in self.cnodes:
            n2 = self.cnodes[n2]

        gnode1 = self.add_node(n1)
        gnode2 = self.add_node(n2)

        self.add_edges(cpt, gnode1, gnode2, size, stretch)

    def add_node(self, n):

        if n not in self:
            self[n] = Gnode(n)
        return self[n]

    def add_edges(self, cpt, gnode1, gnode2, size, stretch):

        gnode1.add_fedge(Gedge(cpt, gnode1, gnode2, size, stretch))
        gnode2.add_redge(Gedge(cpt, gnode2, gnode1, size, stretch))

    @property
    def nodes(self):
        return self.keys()

    @property
    def all_nodes(self):
        return self.cnodes.all_nodes()

    def add_start_nodes(self):
        """Nodes without forward edges are connected to the end node.
        Nodes without reverse edges are connected to the start node."""

        if 'start' in self:
            return

        start = self.add_node('start')
        end = self.add_node('end')

        gnodes = self.values()
        for gnode in gnodes:
            if gnode.redges == [] and gnode != start:
                self.add_edges(None, start, gnode, 0, True)
            if gnode.fedges == [] and gnode != end:
                self.add_edges(None, gnode, end, 0, True)

    def assign_fixed1(self, gnode):

        for edge in gnode.fedges:
            if (not edge.stretch and edge.to_gnode.pos is not None
                    and edge.to_gnode.name != 'end'):
                gnode.pos = edge.to_gnode.pos - edge.size
                return True
        for edge in gnode.redges:
            if (not edge.stretch and edge.to_gnode.pos is not None
                    and edge.to_gnode.name != 'start'):
                gnode.pos = edge.to_gnode.pos + edge.size
                return True
        return False

    def assign_fixed(self, unknown):
        """ Assign node positions to nodes with fixed edge lengths to
        nodes with known positions.  Iterate until no more changes.
        This stage is not needed but provides a minor optimisation."""

        changes = True
        while changes and unknown != []:
            for n in unknown:
                gnode = self[n]
                changes = self.assign_fixed1(gnode)
                if changes:
                    unknown.remove(n)
                    break

    def assign_stretchy1(self, gnode, unknown):

        # For each node, find the longest path between nodes with
        # known positions.  The path between the nodes with known
        # positions is then traversed, measuring the total separation
        # and the number of stretchable components.  (Note, there may
        # be multiple paths of the same length; it does not matter
        # which one is chosen.) The stretch is then calculated from
        # the distance of the path, the total separation, and the
        # number of stretches.  If only one known position is found,
        # the node is dangling and the stretch is zero.

        # TODO, check for multiple paths with a conflict, for example,
        # a stretchy path of a longer distance than a fixed path.

        to_path = self.path_to_closest_known(gnode, forward=True)
        from_path = self.path_to_closest_known(gnode, forward=False)

        to_gnode = to_path.to_gnode
        # Note, from_path is reversed (the edges are backward edges)
        from_gnode = from_path.to_gnode

        if from_gnode.name == 'start' and to_gnode.name == 'end':
            # There is a chance that we have naively picked an
            # unlucky gnode.
            return False

        if from_gnode.name == 'start':
            # Have dangling node, so no stretch needed.
            gnode.pos = to_gnode.pos - to_path.dist
            unknown.remove(gnode.name)
            return True

        if to_gnode.name == 'end':
            # Have dangling node, so no stretch needed.
            gnode.pos = from_gnode.pos + from_path.dist
            unknown.remove(gnode.name)
            return True

        path = self.longest_path(from_gnode, to_gnode)

        # if len(path) == 1:
        #    return False

        # The gnode may not be on this path but this does not matter.
        # It will be processed again later.

        stretches = path.stretches
        separation = to_gnode.pos - from_gnode.pos
        extent = path.dist

        if extent - separation > 1e-6:
            print('Inconsistent %s schematic graph, component(s) will not fit:  separation %s between %s and %s, need %s.\n%s' % (
                self.name, separation, from_gnode, to_gnode, extent, self))

        # This how much each component needs to stretch.
        if stretches == 0:
            stretch = 0
        else:
            stretch = (separation - extent) / stretches
            if stretch < 0:
                stretch = 0

        pos = from_gnode.pos
        for edge in reversed(from_path):
            pos += edge.size
            if edge.stretch:
                pos += stretch

            if edge.from_gnode.pos == None:
                edge.from_gnode.pos = pos
                unknown.remove(edge.from_gnode.name)

        for edge in to_path:
            pos += edge.size
            if edge.stretch:
                pos += stretch

            if edge.to_gnode.pos == None:
                edge.to_gnode.pos = pos
                unknown.remove(edge.to_gnode.name)

        return True

    def assign_stretchy(self, unknown):
        """Use a worklist algorithm to assign nodes with unknown positions
        that are connected via stretchable edges to the known nodes."""

        changes = True
        while changes and unknown != []:
            for n in unknown:
                gnode = self[n]
                if gnode.pos is not None:
                    unknown.remove(n)
                    break

                changes = self.assign_stretchy1(gnode, unknown)
                if changes:
                    self.assign_fixed(unknown)
                    break

    def assign_longest(self, path, unknown):

        if path == []:
            raise RuntimeError('Empty path')

        pos = 0
        for edge in path:
            edge.from_gnode.pos = pos
            pos += edge.size
            unknown.remove(edge.from_gnode.name)

        edge = path[-1]
        edge.to_gnode.pos = pos
        unknown.remove(edge.to_gnode.name)

    def prune(self):
        """Remove redundant paths from graph.  If there are two
        parallel stretchy edges, the longest is chosen."""

        for gnode in self.values():

            def check(edges, grizzle=False):

                # If don't have multiple edges, cannot have redundancy
                if len(edges) < 2:
                    return edges

                fromto = {}
                for edge in edges:
                    key = (edge.from_gnode, edge.to_gnode)
                    if key not in fromto:
                        fromto[key] = []
                    fromto[key].append(edge)

                newedges = []
                for key, edges in fromto.items():

                    best_edge = edges[0]

                    if len(edges) == 1:
                        newedges.append(best_edge)
                        continue

                    fixed = False
                    size = 0
                    for edge in edges:
                        if not edge.stretch:
                            fixed = True
                            size = edge.size
                            best_edge = edge
                            break

                    if fixed and grizzle:
                        for edge in edges:
                            if edge.size != size and not edge.stretch:
                                print('Component fixed size violation for %s, expected %f, got %f' % (
                                    edge.cpt, size, edge.size))
                            elif edge.size > size:
                                print('Component size violation for %s, expected %f or less, got %f' % (
                                    edge.cpt, size, edge.size))

                    elif not fixed:
                        # Choose largest component size since all are stretchy
                        for edge in edges:
                            if edge.size > size:
                                size = edge.size
                                best_edge = edge

                    newedges.append(best_edge)

                return newedges

            gnode.fedges = check(gnode.fedges, grizzle=True)
            gnode.redges = check(gnode.redges)

    def suggest_edges(self):

        for node in self.values():
            fedges = node.fedges
            redges = node.redges
            if len(redges) != 0:
                continue
            if len(fedges) != 2:
                continue
            fedge1 = fedges[0]
            fedge2 = fedges[1]

            node1 = fedge1.to_gnode
            node2 = fedge2.to_gnode
            if len(node1.fedges) != 2:
                continue
            if len(node2.fedges) != 2:
                continue
            if node1.fedges[0].to_gnode == node2.fedges[0].to_gnode:
                node3 = node1.fedges[0].to_gnode
                fedge3 = node1.fedges[0]
                fedge4 = node2.fedges[0]
            elif node1.fedges[0].to_gnode == node2.fedges[1].to_gnode:
                node3 = node1.fedges[0].to_gnode
                fedge3 = node1.fedges[0]
                fedge4 = node2.fedges[1]
            elif node1.fedges[1].to_gnode == node2.fedges[0].to_gnode:
                node3 = node1.fedges[1].to_gnode
                fedge3 = node1.fedges[1]
                fedge4 = node2.fedges[0]
            elif node1.fedges[1].to_gnode == node2.fedges[1].to_gnode:
                node3 = node1.fedges[1].to_gnode
                fedge3 = node1.fedges[1]
                fedge4 = node2.fedges[1]
            else:
                continue
            if len(node3.fedges) != 0:
                continue
            if fedge1.size == fedge3.size and fedge2.size == fedge4.size:
                pass
            elif fedge1.size == fedge2.size and fedge3.size == fedge4.size:
                pass
            else:
                continue

            print('Suggestion: add a constraint between nodes %s and %s for %s graph' %
                  (node1, node2, self.name))

    def solve(self, stage=None):
        """
        Solve graph assigning gnode positions.

        stage 0 --- do nothing
        stage 1 --- prune redundant edges
        stage 2 --- add start/end nodes and assign gnode positions on longest path
        stage 3 --- assign gnode positions with fixed positions to known gnodes
        stage 4 --- assign all gnode positions
        """

        unknown = list(self.keys())

        if unknown == []:
            # If there are no constraints, all the nodes have distance 0.
            pos = {}
            for n, gnode in self.cnodes.items():
                pos[n] = 0
            return pos, 0

        if stage == 0:
            return

        # Prune redundant edges from graph.
        self.prune()

        self.suggest_edges()

        if stage == 1:
            return

        self.add_start_nodes()
        unknown.append('start')
        unknown.append('end')

        # Find longest path through the graph.  This provides the
        # dimension for the graph.
        path = self.longest_path(self['start'], self['end'])

        # Assign gnodes on the longest path.
        self.assign_longest(path, unknown)

        if stage == 2:
            return

        # Assign gnodes at a fixed distance from nodes with known positions.
        self.assign_fixed(unknown)

        if stage == 3:
            return

        # Assign remaining gnodes.
        self.assign_stretchy(unknown)

        if unknown != []:
            raise ValueError('Cannot assign nodes %s for %s graph.\n%s' %
                             (unknown, self.name, self))

        self.check_positions()

        try:
            pos = {}
            for n, gnode in self.cnodes.items():
                pos[n] = self[gnode].pos

        except KeyError:
            # TODO determine which components are not connected.
            msg = "The %s schematic graph is dodgy, probably a component is attached to the wrong nodes.\n" % self.name
            if self.debug:
                msg += str(self)
            raise AttributeError(msg)

        distance_max = self['end'].pos

        return pos, distance_max

    def longest_path(self, from_gnode, to_gnode):
        """Find longest path through DAG from from_gnode to to_gnode
        or to a gnode with a known position."""

        # gnode.dist specifies the distance to the to_gnode.  It is
        # None if the distance has yet to be determined and is -1 if
        # the to_gnode cannot be reached from gnode.
        for gnode in self.values():
            gnode.dist = None
            gnode.next = None

        def traverse(gnode, depth=0):

            if depth > 1000:
                raise RuntimeError('Recursion depth exceeded')

            # If have visited this gnode before then know dist.
            if gnode.dist is not None:
                return gnode.dist

            # If have reached goal, return dist of 0.
            if gnode == to_gnode:
                gnode.dist = 0
                return 0

            # Do not traverse nodes at known positions.
            if gnode.pos is not None and gnode != from_gnode:
                return -1

            # Mark node as visited but distance as unknown
            gnode.dist = -1

            edges = gnode.fedges
            for edge in edges:
                dist = traverse(edge.to_gnode, depth + 1)
                if dist >= 0 and gnode.dist < dist + edge.size:
                    gnode.dist = dist + edge.size
                    # Mark current best edge
                    gnode.next = edge

            return gnode.dist

        try:
            traverse(from_gnode)
        except RuntimeError:
            msg = "The %s schematic graph is dodgy, probably a component is attached to the wrong nodes.\n" % self.name
            if self.debug:
                msg += str(self)

            raise RuntimeError(msg)

        return self.makepath(from_gnode)

    def makepath(self, from_gnode):

        # Construct path
        path = GraphPath()
        gnode = from_gnode
        while gnode.next is not None:
            if gnode.next.to_gnode == gnode:
                msg = "The %s schematic graph is dodgy, probably a component is attached to the wrong nodes or a direction attribute is bogus.\n" % self.name
                if self.debug:
                    msg += str(self)
                raise RuntimeError(msg)

            path.append(gnode.next)
            gnode = gnode.next.to_gnode

        return path

    def path_to_closest_known(self, from_gnode, forward=True):
        """Find path through DAG to the closest node with a known pos."""

        def traverse(gnode, depth=0):
            if depth > 1000:
                raise RuntimeError('Recursion depth exceeded')

            if gnode.name in ('start', 'end'):
                # Choose as last resort
                gnode.next = None
                return 1000

            if gnode.pos is not None:
                gnode.next = None
                return gnode.pos

            edges = gnode.fedges if forward else gnode.redges
            min_dist = 2000
            for edge in edges:
                next_gnode = edge.to_gnode
                # TODO, memoize
                dist = traverse(next_gnode, depth + 1) - edge.size
                if dist < min_dist:
                    min_dist = dist
                    next_gnode.prev = edge
                    gnode.next = edge
            return min_dist

        from_gnode.dist = 0
        try:
            traverse(from_gnode)
        except RuntimeError:
            # W a b; right
            # W c d; right
            # W a c; down
            # W b d; up
            above = 'above' if self.name == 'vertical' else 'left of'
            below = 'below' if self.name == 'vertical' else 'right of'
            msg = "The %s schematic graph has a loop.  For example, a node needs to be both %s and %s another node.  Probably a component is attached to the wrong nodes.\n" % \
                (self.name, above, below)
            if self.debug:
                msg += str(self)
            raise RuntimeError(msg)

        return self.makepath(from_gnode)

    def check_positions(self):

        for gnode in self.values():
            for edge in gnode.fedges:
                dist = edge.to_gnode.pos - gnode.pos
                if edge.stretch:
                    if dist - edge.size < -1e-6:
                        print('Distance conflict %s vs %s in %s graph for %s between nodes %s and %s,'
                              ' due to incompatible sizes' % (
                                  dist, edge.size, self.name,
                                  edge.name, gnode.fmt_name,
                                  edge.to_gnode.fmt_name))
                else:
                    if abs(dist - edge.size) > 1e-6:
                        print('Stretch conflict %s vs %s in %s graph for %s between nodes %s and %s,'
                              ' due to incompatible sizes' % (
                                  dist, edge.size, self.name,
                                  edge.name, gnode.fmt_name,
                                  edge.to_gnode.fmt_name))

    def dot(self, filename=None, stage=None, tweak_node_label=True):
        """Generate directed graph using graphviz notation"""

        from .system import tmpfilename, run_dot
        from .schematic import display_matplotlib
        from os import path

        def fmt_dec(value):
            return ('%.2f' % value).rstrip('0').rstrip('.')

        def fmt_node_label(node):

            if isinstance(node.name, tuple):
                label = ', '.join(node.name)
            else:
                label = node.name

            # Hack since xlabel not supported bu dot2tex.
            if tweak_node_label and node.pos is not None:
                label += ' @' + fmt_dec(node.pos)

            return label

        if filename is None:
            filename = tmpfilename('.png')
            self.dot(filename=filename, stage=stage)
            display_matplotlib(filename)
            return

        base, ext = path.splitext(filename)
        if ext in ('.pdf', '.png'):
            dotfilename = filename + '.dot'
            self.dot(dotfilename, stage=stage)
            run_dot(dotfilename, filename)
            return

        self.solve(stage=stage)

        dotfile = open(filename, 'w')
        dotfile.write('digraph {\n\tgraph [rankdir=LR];\n')

        for gnode in self.values():
            if hasattr(gnode, 'fixed'):
                colour = colours['fixednode']
            else:
                colour = colours['assignednode'] if gnode.pos is not None else colours['unassignednode']
            if gnode.name == 'start':
                colour = colours['startnode']
            elif gnode.name == 'end':
                colour = colours['endnode']

            pos = gnode.pos
            if pos is None or pos < 1e-6:
                pos = 0

            dotfile.write('\t"%s"\t [style=filled, color=%s];\n' % (
                fmt_node_label(gnode), colour))

        for gnode in self.values():
            for edge in gnode.fedges:
                colour = colours['stretchedge'] if edge.stretch else colours['fixededge']
                dotfile.write('\t"%s" ->\t"%s" [ color=%s, label="%s%s" ];\n' % (
                    fmt_node_label(gnode), fmt_node_label(
                        edge.to_gnode), colour,
                    fmt_dec(edge.size), '*' if edge.stretch else ''))

        dotfile.write('}\n')
        dotfile.close()



"""This module provides the Units class for simplification of units.
It should be rolled into SymPy.  It can perform simplification of
units, e.g., volts / amperes -> ohms.

Copyright 2020--2021 Michael Hayes, UCECE

"""

import sympy.physics.units as u
from sympy.physics.units.systems.si import dimsys_SI
from sympy.physics.units.systems import SI
from sympy.physics.units import UnitSystem, Quantity
from sympy import S, __version__ as sympy_version


dB = Quantity('dB', 'dB')


class Units(object):

    def __init__(self, unit_system="SI"):

        self.unit_system = UnitSystem.get_unit_system(unit_system)
        self.dim_sys = self.unit_system.get_dimension_system()
        self._mapping = {}

        for i in u.__dict__:
            unit = getattr(u, i)
            if not isinstance(unit, u.Quantity):
                continue

            key = self._makekey(unit)

            # Use earlier defined units

            if key not in self._mapping:
                self._mapping[key] = unit

        # Remove entry for no units.
        self._mapping.pop(self._makekey(1))

        # Add entry for S * ohm, etc.
        key = (None, ) * len(key)
        self._mapping[key] = S.One

    def _get_dependencies(self, unit):

        dim = self.unit_system.get_dimensional_expr(unit)
        return self.dim_sys.get_dimensional_dependencies(dim)

    def _makekey(self, unit):

        deps = self._get_dependencies(unit)

        if sympy_version >= '1.11':
            key = tuple([deps.get(u.Dimension(str(dim.name)))
                         for dim in self.dim_sys.base_dims])
        else:
            key = tuple([deps.get(str(dim.name))
                         for dim in self.dim_sys.base_dims])
        return key

    def simplify_units(self, unit):

        key = self._makekey(unit)
        if not key in self._mapping:
            return unit
        result = self._mapping[key]

        # V s or Wb?  In the context of Laplace transforms, V s makes more
        # sense since the Laplace domain voltage has units (V / rad / s).
        # However, for magnetic field strength, Wb makes more sense.  Since
        # this is for circuit analysis we plump for V s.
        if result.has(u.webers):
            result = result.replace(u.webers, u.volt * u.s)

        # There are probably many more special cases like this.
        if result == u.hbar:
            result = u.joule / u.Hz

        if not unit.has(u.rad):
            return result

        # If original expression has rad (or 1 / rad) then this will
        # get lost in the mapping, so need to reapply it.
        factors = unit.as_ordered_factors()
        if u.rad in factors:
            return result * u.rad
        return result / u.rad

    def simplify(self, expr):

        value, unit = self.as_value_unit(expr)
        return value * self.simplify_units(unit)

    def as_value_unit(self, expr):
        return as_value_unit(expr)


def as_value_unit(expr):

    if isinstance(expr, u.Quantity):
        return 1, expr

    if not expr.has(u.Quantity):
        return expr, 1

    if expr.is_Pow and expr.args[1] == -1:
        value, unit = as_value_unit(expr.args[0])
        return S.one / value, S.one / unit

    defs = {x: 1 for x in expr.args if not x.has(u.Quantity)}
    unit = expr.subs(defs)

    value = expr / unit
    if value.has(u.Quantity):
        # FIXME: This function only works for something like 42 * volt or 42 * amp * ohm.
        # It fails for 4 * amp * 2 * ohm + 42 * volt.
        raise ValueError('Expression not of form value * units: %s' % expr)

    return value, unit


units = Units()
volts = u.volts
amperes = u.amperes
ohms = u.ohms
siemens = u.siemens
watts = u.watts



"""This module provides support for z transforms.

Copyright 2020--2023 Michael Hayes, UCECE

"""

from .transformer import UnilateralForwardTransformer
from .ratfun import Ratfun
from .sym import sympify, simplify, miscsymbol, AppliedUndef
from .utils import factor_const, scale_shift
from .extrafunctions import UnitImpulse, UnitStep, dtrect
import sympy as sym
from sympy.simplify.fu import TR6, TR9

__all__ = ('ZT', )


# Check the structure of the given expression
def is_multiplied_with(expr, n, cmp, ret):

    if expr.is_Piecewise:
        return False

    ret_flag = False
    # Check for multiplication  with n
    if cmp == 'n' and expr == n:  # only n
        ret += [n]
        ret_flag = True
    elif (cmp == 'n' and expr.is_Pow and expr.args[0] == n and
          expr.args[1].is_integer and expr.args[1] >= 1):     # only n**i
        ret += [n]
        ret_flag = True
    elif cmp == 'n' and expr.is_Mul:   # multiplication with n
        for i in range(len(expr.args)):
            if (expr.args[i].is_Pow and expr.args[i].args[0] == n and
                    expr.args[i].args[1].is_integer and expr.args[i].args[1] >= 1):
                ret += [n]
                ret_flag = True
                break
            elif expr.args[i] == n:
                ret += [n]
                ret_flag = True
                break

    # Check for multiplication with a**(b*n+c)
    elif (cmp == 'a**n' and expr.is_Pow and
          ((expr.args[1]).as_poly(n)).is_linear and (not expr.args[0].has(n))):
        ret += [expr]
        ret_flag = True
    elif cmp == 'a**n' and expr.is_Mul:
        for i in range(len(expr.args)):
            if ((expr.args[i].is_Pow and
                 ((expr.args[i].args[1]).as_poly(n)).is_linear and
                 (not expr.args[i].args[0].has(n)))):
                ret += [expr.args[i]]
                ret_flag = True
                break

    # Check for multiplication with exp(b*n+c)
    elif (cmp == 'exp(n)' and len(expr.args) == 1 and expr.is_Function and
          ((expr.args[0]).as_poly(n)).is_linear and expr.func == sym.exp):
        ret += [expr]
        ret_flag = True
    elif cmp == 'exp(n)' and expr.is_Mul:
        for i in range(len(expr.args)):
            if (expr.args[i].is_Function and expr.args[i].func == sym.exp and
                    ((expr.args[i].args[0]).as_poly(n)).is_linear):
                ret += [expr.args[i]]
                ret_flag = True
                break

    # Check for multiplication with u(n-n0)
    elif (cmp == 'UnitStep' and len(expr.args) == 1 and expr.is_Function and  # step only
          ((expr.args[0]).as_poly(n)).is_linear and expr.func in (sym.Heaviside, UnitStep)):
        ret += [expr]
        ret_flag = True
    elif cmp == 'UnitStep' and expr.is_Mul:
        for i in range(len(expr.args)):
            if expr.args[i].is_Function and expr.args[i].func in (sym.Heaviside, UnitStep) and ((expr.args[i].args[0]).as_poly(n)).is_linear:
                ret += [expr.args[i]]
                ret_flag = True
                break

    # Check for multiplication with dtrect
    elif (cmp == 'rect' and len(expr.args) == 1 and expr.is_Function and
          expr.func == dtrect and ((expr.args[0]).as_poly(n)).is_linear):
        ret += [expr]
        ret_flag = True
    elif cmp == 'rect' and expr.is_Mul:
        for i in range(len(expr.args)):
            if expr.args[i].is_Function and expr.args[i].func == dtrect and ((expr.args[i].args[0]).as_poly(n)).is_linear:
                ret += [expr.args[i]]
                ret_flag = True
                break

   # Check for multiplication with sin
    elif (cmp == 'sin(n)' and len(expr.args) == 1 and expr.is_Function and  # sin() only
          ((expr.args[0]).as_poly(n)).is_linear and expr.func == sym.sin):
        ret += [expr]
        ret_flag = True
    elif (cmp == 'sin(n)' and expr.is_Pow and expr.args[0].func == sym.sin and
          (expr.args[0].args[0]).as_poly(n).is_linear):
        ret += [expr.args[0]]
        ret_flag = True
    elif cmp == 'sin(n)' and expr.is_Mul:
        for i in range(len(expr.args)):
            if (expr.args[i].is_Function and expr.args[i].func == sym.sin and
                    ((expr.args[i].args[0]).as_poly(n)).is_linear):
                ret += [expr.args[i]]
                ret_flag = True
                break
            elif (expr.args[i].is_Pow and expr.args[i].args[0].func == sym.sin and
                  ((expr.args[i].args[0].args[0]).as_poly(n)).is_linear):
                ret += [expr.args[i].args[0]]
                ret_flag = True
                break

    # Check for multiplication with cos
    elif (cmp == 'cos(n)' and len(expr.args) == 1 and expr.is_Function and  # cos only
          ((expr.args[0]).as_poly(n)).is_linear and expr.func == sym.cos):
        ret += [expr]
        ret_flag = True
    elif cmp == 'cos(n)' and expr.is_Pow and expr.args[0].func == sym.cos and (expr.args[0].args[0]).as_poly(n).is_linear:
        ret += [expr.args[0]]
        ret_flag = True
    elif cmp == 'cos(n)' and expr.is_Mul:
        for i in range(len(expr.args)):
            if (expr.args[i].is_Function and expr.args[i].func == sym.cos and
                    ((expr.args[i].args[0]).as_poly(n)).is_linear):
                ret += [expr.args[i]]
                ret_flag = True
                break
            elif (expr.args[i].is_Pow and expr.args[i].args[0].func == sym.cos and
                  ((expr.args[i].args[0].args[0]).as_poly(n)).is_linear):
                ret += [expr.args[i].args[0]]
                ret_flag = True
                break

    return ret_flag


class ZTransformer(UnilateralForwardTransformer):

    name = 'z-transform'

    def noevaluate(self, expr, n, z):

        foo = expr * z**(-n)
        result = sym.Sum(foo, (n, 0, sym.oo))
        return result

    def rewrite(self, expr, var):
        # This is needed to handle expressions like (2*n + 3)**2
        return sym.expand(expr)

    def check(self, expr, n, z, **kwargs):

        if expr.has(z):
            self.error('Expression depends on z')

    def key(self, expr, n, z, **kwargs):
        return expr, n, z,

    def func(self, expr, n, z):

        if not isinstance(expr, AppliedUndef):
            self.error('Expecting function')

        scale, shift = scale_shift(expr.args[0], n)

        # Convert v(n) to V(z), etc.
        name = expr.func.__name__
        func = sym.Function(name[0].upper() + name[1:])

        if not scale.is_constant():
            self.error('Cannot determine if time-expansion or decimation')

        if scale == 1:
            result = func(z)

            if shift != 0:
                # x[ n+i ] = z^i X(z) -z^i x(0) - .... - z x(i-1)
                if shift.is_positive:
                    result = result * z ** shift
                    for ind in range(shift):
                        result -= expr.func(ind) * z ** (shift - ind)
                # x[ n-i ] = z^(-i) X(z)
                else:
                    result = result * z ** shift
            return result

        if scale.is_integer:
            # Down-sampling produces aliasing
            # Sum(X(z**(1 / M) * exp(-j * 2 * pi * m / M), (m, 0, M - 1)) / M)
            # Down-sampling is not shift invariant so z-transform
            # is an approximation.
            m = self.dummy_var(expr, 'm', level=0, real=True)
            expr = func(z**(1 / scale) *
                        sym.exp(-sym.I * 2 * sym.pi * m / scale))
            return sym.Sum(expr, (m, 0, scale - 1)) / scale

        if not scale.is_rational:
            self.error('Cannot handle arbitrary scaling')

        if scale.p != 1:
            self.error('Cannot handle non-integer time-expansion')

        result = func(z ** scale.q)

        if shift != 0:
            result = result * z ** shift
        return result

    def sum(self, expr, n, z):

        const, expr = factor_const(expr, n)

        if len(expr.args) != 2:
            self.error('Cannot compute z-transform')

        if not isinstance(expr, sym.Sum):
            self.error('Cannot compute z-transform')

        # Look for integration of function
        if (isinstance(expr.args[0], AppliedUndef)
            and expr.args[1][0] == expr.args[0].args[0]
            and expr.args[1][1] == -sym.oo
                and expr.args[1][2] == n):
            return self.func(expr.args[0].subs(expr.args[0].args[0], n), n, z) / (1 - 1 / z)

        # Look for convolution sum
        var = expr.args[1][0]
        if (expr.args[1][1] != -sym.oo) or (expr.args[1][2] != sym.oo):
            self.error('Need indefinite limits')

        expr = expr.args[0]
        if ((len(expr.args) != 2)
            or (not isinstance(expr.args[0], AppliedUndef))
                or (not isinstance(expr.args[1], AppliedUndef))):
            self.error('Need convolution of two functions')

        f1 = expr.args[0]
        f2 = expr.args[1]

        # TODO: apply similarity theorem if have f(a tau) etc.

        if ((f1.args[0] != var or f2.args[0] != n - var)
                and (f2.args[0] != var or f1.args[0] != n - var)):
            self.error('Cannot recognise convolution')

        name = f1.func.__name__
        func1 = sym.Function(name[0].upper() + name[1:])

        name = f2.func.__name__
        func2 = sym.Function(name[0].upper() + name[1:])

        F1 = func1(z)
        F2 = func2(z)

        return F1 * F2

    def term(self, expr, n, z):
        from .sym import dt

        # Unilateral ZT ignores expr for n < 0 so remove Piecewise.
        if expr.is_Piecewise:
            if (expr.args[0].args[1].func is sym.GreaterThan and
                    expr.args[0].args[1].args[1] == 0):
                const, e = factor_const(expr.args[0].args[1].args[0], n)
                if e is n:
                    expr = expr.args[0].args[0]

        const, expr = factor_const(expr, n)

        expr = expr.replace(UnitStep(n * dt), 1)

        if expr.has(sym.Sum):
            try:
                return self.sum(expr, n, z) * const
            except:
                pass

        nsym = sympify(str(n))
        expr = expr.replace(nsym, n)

        # foo = 1 / (1 - 1 / z)
        # factors = expr.as_ordered_factors()
        # if foo in factors:
        #     could remove factor, find ZT of rest, and then integrate....

        if expr.has(AppliedUndef):

            rest = sym.S.One
            expr = expr.cancel()
            for factor in expr.as_ordered_factors():
                if isinstance(factor, AppliedUndef):
                    result = self.func(factor, n, z)
                else:
                    if factor.has(n):
                        self.error('TODO: need derivative of undefined'
                                   ' function for %s' % factor)
                    rest *= factor
            return result * rest * const

        invz = z ** -1

        result = None
        args = expr.args
        xn_fac = []

        if expr.is_Function and expr.func == UnitImpulse:
            if args[0] is n:
                result = 1
            delay = n - args[0]
            if not delay.has(n):
                result = invz ** delay

        elif expr == 1:
            # Unilateral ZT
            result = 1 / (1 - invz)

        elif (expr.is_Function and
              expr.func in (sym.Heaviside, UnitStep, sym.sign)):
            if args[0] is n:
                result = 1 / (1 - invz)
            else:
                delay = n - args[0]
                if not delay.has(n):
                    result = invz ** delay * 1 / (1 - invz)

        # sin(b*n+c)
        elif (expr.is_Function and expr.func == sym.sin and (args[0].as_poly(n)).is_linear):
            bb = args[0].coeff(n, 1)
            cc = args[0].coeff(n, 0)
            result = (sym.sin(cc) + sym.sin(bb - cc) * invz) / \
                (1 - 2 * sym.cos(bb) * invz + invz**2)
            result = sym.simplify(result)

        # cos(b*n+c)
        elif (expr.is_Function and expr.func == sym.cos and (args[0].as_poly(n)).is_linear):
            bb = args[0].coeff(n, 1)
            cc = args[0].coeff(n, 0)
            result = (sym.cos(cc) - sym.cos(bb - cc) * invz) / \
                (1 - 2 * sym.cos(bb) * invz + invz**2)
            result = sym.simplify(result)

        # Multiplication with n       use n*x(n)  o--o  -z d/dz X(z)
        elif is_multiplied_with(expr, n, 'n', xn_fac):
            expr = expr / xn_fac[0]
            X = self.term(expr, n, z)
            result = sym.simplify(-z * sym.diff(X, z))

        # Multiplication with a**(b*n+c)   use    lam**n *x(n)  o--o  X(z/lam)
        elif is_multiplied_with(expr, n, 'a**n', xn_fac):
            expr /= xn_fac[0]
            ref = xn_fac[0].args
            lam = ref[0]
            bb = ref[1].coeff(n, 1)
            cc = ref[1].coeff(n, 0)
            X = self.term(expr, n, z)
            result = lam**cc * sym.simplify(X.subs(z, z / lam**bb))

        # Multiplication with exp(b*n+c)   use    exp**n *x(n)  o--o  X(z/exp(1))
        elif is_multiplied_with(expr, n, 'exp(n)', xn_fac):
            expr /= xn_fac[0]
            ref = xn_fac[0].args
            bb = ref[0].coeff(n, 1)
            cc = ref[0].coeff(n, 0)
            X = self.term(expr, n, z)
            result = sym.exp(cc) * sym.simplify(X.subs(z, z / sym.exp(bb)))

        # Multiplication with u(n-n0) * x(n)    o--o     X(z)- (x[0] - x[1]/z - .. - x[n0-1]/z**(n0-1)
        elif is_multiplied_with(expr, n, 'UnitStep', xn_fac):
            expr /= xn_fac[0]
            delay = n - xn_fac[0].args[0]
            X = self.term(expr, n, z)
            sum_X = 0
            for ii in range(delay):
                sum_X -= expr.subs(n, ii) * invz**ii
            result = X + sum_X

        if result is None:
            # Use m instead of n to avoid n and z in same expr.
            # TODO, check if m already used...
            msym = sympify('m', real=True)
            nsym = sympify(str(n))
            zsym = sympify(str(z))
            result = sym.Sum(expr.subs(nsym, msym) *
                             zsym**msym, (msym, 0, sym.oo))

        return const * result


ztransformer = ZTransformer()


def ZT(expr, n, z, evaluate=True, **kwargs):
    """Compute unilateral Z-Transform transform of expr with lower limit 0.

    Undefined functions such as v[n] are converted to V(z)."""

    return ztransformer.transform(expr, n, z,
                                  evaluate=evaluate, **kwargs)


def ztransform(expr, n, z, evaluate=True, **kwargs):
    """Compute unilateral Z-Transform transform of expr with lower limit 0.

    Undefined functions such as v[n] are converted to V(z)."""

    return ztransformer.transform(expr, n, z,
                                  evaluate=evaluate, **kwargs)


from .expr import Expr  # nopep8



"""
This module performs schematic drawing using circuitikz from a netlist::

    >>> from lcapy import Schematic
    >>> sch = Schematic('''
    ... P1 1 0.1; down
    ... R1 3 1; right
    ... L1 2 3; right
    ... C1 3 0; down
    ... P2 2 0.2; down
    ... W 0 0.1; right
    ... W 0.2 0.2; right''')
    >>> sch.draw()

Copyright 2014--2023 Michael Hayes, UCECE
"""

# Strings starting with ;; are schematic options.  They are parsed in
# netfile.py and added to the opts attribute of the netlist.  They get
# passed to the draw method of the Schematic class.


from __future__ import print_function
from .config import autoground_default
from .latex import latex_format_label
from .expr import Expr
from . import schemcpts
import sympy as sym
from .engformatter import EngFormatter
from .schemmisc import Pos
from .schemnode import Node
from .schemplacer import schemplacer
from .opts import Opts
from .netfile import NetfileMixin
from .system import LatexRunner, PDFConverter, tmpfilename
from .nodes import parse_nodes
from os import path, remove
from collections import OrderedDict
from warnings import warn


__all__ = ('Schematic', )

# Default dots per inch for png files
PNG_DPI = 300

CIRCUITIKZ_MIN_VERSION = '1.4.5'


def display_matplotlib(filename, dpi=PNG_DPI, scale=2):

    from matplotlib.pyplot import figure
    from matplotlib.image import imread

    img = imread(filename)
    h, w, d = img.shape
    width = scale * w / dpi
    height = scale * h / dpi
    fig = figure(figsize=(width, height))
    ax = fig.add_subplot(111)
    ax.imshow(img)
    ax.axis('off')
    fig.subplots_adjust(left=0.05, right=0.95, top=0.95, bottom=0.05)
    fig.show()


def png_image_size(filename):

    import struct

    with open(filename, 'rb') as f:
        header = f.read(25)

    if (header[:8] != b'\211PNG\r\n\032\n' and (header[12:16] != b'IHDR')):
        raise Exception('%s not a png image' % filename)
    w, h = struct.unpack('>LL', header[16:24])
    width = int(w)
    height = int(h)
    return width, height


class SchematicOpts(Opts):

    def __init__(self):

        super(SchematicOpts, self).__init__(
            {'draw_nodes': 'primary',
             'label_values': True,
             'label_ids': True,
             'annotate_values': False,
             'label_nodes': 'primary',
             'anchor': 'south east',
             'autoground': 'none',
             'scale': 1.0,
             'dpi': PNG_DPI,
             'cpt_size': 1.5,
             'node_spacing': 2.0,
             'help_lines': 0.0,
             'style': 'american',
             'voltage_dir': 'RP'})


class Schematic(NetfileMixin):

    def __init__(self, filename=None, allow_anon=False, **kwargs):

        self.elements = OrderedDict()
        self.nodes = {}
        self.hints = False
        self._init_parser(schemcpts, allow_anon)
        self.cpt_size = 1.2
        self.node_spacing = 2.0
        self.scale = 1.0
        self.dummy_node = 0
        self.context = None
        self.debug = 0
        self.defines = {}
        self.node_positions = {}

        if filename is not None:
            self.netfile_add(filename)

    def __repr__(self):

        return self.netlist()

    def __getitem__(self, name):
        """Return component by name"""
        try:
            return self.elements[name]
        except KeyError:
            raise AttributeError('Unknown component %s' % name)

    def netfile_add(self, filename):
        """Add the nets from file with specified filename"""

        self._netfile_add(filename)

    def add(self, string):
        """Add a component to the netlist.
        The general form is: 'Name Np Nm args'
        where Np is the positive node and Nm is the negative node.

        A positive current is defined to flow from the positive node
        to the negative node.
        """

        self._add(string)

    def netlist(self):
        """Return the current netlist"""

        return '\n'.join([elt.__str__() for elt in self.elements.values()])

    def _node_add(self, node_name, elt, auxiliary=False):

        if node_name not in self.nodes:
            self.nodes[node_name] = Node(node_name)
        node = self.nodes[node_name]

        node.append(elt)

        if node.auxiliary is None:
            node.auxiliary = auxiliary
        elif node.auxiliary and not auxiliary:
            # An auxiliary node can be made a proper node.
            node.auxiliary = False

        return node

    def _format_name(self, cpt_type, cpt_id):

        name = cpt_type
        subscript = cpt_id[1:] if cpt_id.startswith('_') else cpt_id

        if cpt_type == 'REL':
            name = r'\mathcal{R}'
        elif len(name) > 1:
            name = r'\mathrm{%s}' % name

        if subscript != '':
            if len(subscript) > 1:
                subscript = r'\mathrm{%s}' % subscript
            name = name + '_{%s}' % subscript
        return latex_format_label('$' + name + '$')

    def _format_expr(self, expr):

        return Expr(expr, cache=False).latex_math()

    def _format_value_units(self, value, units):

        return EngFormatter().latex_math(value, units)

    def _cpt_add(self, cpt):

        if cpt.offset != 0 and len(cpt.node_names) == 2:
            # Change component and add wires.

            n1 = cpt.node_names[0]
            n2 = cpt.node_names[1]
            # If have only two parallel components, can used fixed node names.
            self.dummy_node += 1
            on1 = n1 + 'off%d' % self.dummy_node
            self.dummy_node += 1
            on2 = n2 + 'off%d' % self.dummy_node

            angle = 90
            size = cpt.offset
            if size < 0:
                size = -size
                angle = -angle

            w1 = 'W %s %s; rotate=%s, size=%s' % (
                n1, on1, cpt.angle + angle, size)
            w2 = 'W %s %s; rotate=%s, size=%s' % (
                n2, on2, cpt.angle + angle, size)

            self.add(w1)
            self.add(w2)

            # Add open-circuit to ensure alignment.
            o = 'O %s %s; rotate=%s, size=%s' % (n1, n2, cpt.angle, cpt.size)
            self.add(o)

            # Rename nodes
            parts = cpt.net.split(' ')
            parts[1] = on1
            parts[2] = on2
            net = ' '.join(parts)
            cpt.opts.strip('offset')
            self.add('%s; %s' % (net, cpt.opts))
            return

        # There are two possible labels for a component:
        # 1. Component name, e.g., R1
        # 2. Component value, expression, or symbol

        id_label = self._format_name(cpt.type, cpt.id)
        value_label = None

        if cpt.type in ('A', 'O', 'P', 'W') or id_label.find('#') != -1:
            id_label = None

        if cpt.type in ('A', 'S', 'SW', 'U'):
            value_label = ''

        unify = False
        if cpt.args != ():

            # TODO, extend for mechanical and acoustical components.
            units_map = {'V': 'V', 'I': 'A', 'R': '$\Omega$',
                         'C': 'F', 'L': 'H'}

            expr = cpt.args[0]
            if cpt.classname in ('Vstep', 'Istep'):
                expr = '(%s) * Heaviside(t)' % expr
                value_label = self._format_expr(expr)
            elif cpt.classname in ('Vs', 'Is'):
                value_label = self._format_expr(expr)
            elif cpt.classname == 'TF':
                expr = sym.sympify(expr)
                if expr.is_Pow and expr.args[1] == -1:
                    value_label = '%s:1' % (1 / expr)
                else:
                    value_label = '1:%s' % expr
            elif cpt.type in ('F', 'H') and len(cpt.args) > 1:
                # This is hard to give a reasonable label since the
                # control current is specified by a voltage source.
                # The user will have to override manually.
                expr = cpt.args[1]
                value_label = self._format_expr(expr)
            elif cpt.classname not in ('TP',):
                try:
                    # Handle things like 9/1000 that can occur
                    # when substituting cpt values.
                    value = float(sym.Rational(expr))
                    if cpt.type in units_map:
                        value_label = self._format_value_units(
                            value, units_map[cpt.type])
                    else:
                        value_label = self._format_expr(expr)

                except (ValueError, TypeError):
                    # This catches non numeric arg.
                    value_label = self._format_expr(expr)

            # Ensure labels are the same when the value is not specified.
            # This will prevent printing the name and value.
            unify = expr == cpt.type + cpt.id

            # Currently, we only annnotate the component with the
            # value, expression, or symbol.  If this is not specified,
            # it defaults to the component identifier.  Note, there
            # are some objects we do not want to label, such as wires
            # and ports.
        cpt.id_label = '' if id_label is None else latex_format_label(id_label)
        cpt.value_label = cpt.id_label if value_label is None \
            else latex_format_label(value_label)

        if unify:
            cpt.value_label = cpt.id_label

        if cpt.opts_string != '':
            self.hints = True

        if cpt.name in self.elements:
            warn('Overriding component %s' % cpt.name)
            # Need to search lists and update component.

        if cpt.type == 'XX' and 'nodes' in cpt.opts:
            self.positions = parse_nodes(cpt.opts['nodes'])

        self.elements[cpt.name] = cpt

        # Perhaps don't show nodes if cpt invisible?
        if not cpt.ignore:

            # The auxiliary nodes are for behind the scenes
            # use, such as finding the component centre.
            for node_name in cpt.auxiliary_node_names:
                pinname = node_name.split('.')[-1]
                if pinname in cpt.required_auxiliary:
                    self._node_add(node_name, cpt, auxiliary=True)

            # The required nodes are those explicitly
            # specified in the schematic.
            for node in cpt.required_node_names:
                self._node_add(node, cpt, auxiliary=False)

    def _setup(self, autoground):
        # This is called before node positions are assigned.

        if autoground in ('true', 'True', True):
            autoground = autoground_default
        elif autoground in (False, None, 'false', 'none'):
            autoground = False

        for elt in self.elements.values():
            elt.parse_nodes()

        for elt in self.elements.values():
            elt.check_nodes()

        if autoground:
            for elt in self.elements.values():
                elt.autoground(autoground)

        for elt in self.elements.values():
            elt.setup()

        # Cannot split nodes until know all the nodes.
        for elt in self.elements.values():
            elt.process_implicit_nodes()

    def make_graphs(self, debug=0, **kwargs):

        # This is only used by schtex for debugging
        autoground = kwargs.get('autoground', False)
        self._setup(autoground)
        placer = schemplacer(self.elements, self.nodes, 'graph', debug)
        placer._make_graphs()
        return placer.xgraph, placer.ygraph

    def _positions_calculate(self, method='graph', debug=False, **kwargs):

        autoground = kwargs.get('autoground', False)
        self._setup(autoground)

        if self.node_positions != {}:
            for k, v in self.node_positions.items():
                # Opamp ground node not included
                if k in self.nodes:
                    self.nodes[k].pos = v

            nodes = [node for node in self.nodes.values() if not node.is_split]

            x = [node.pos.x for node in nodes if not node.auxiliary]
            y = [node.pos.y for node in nodes if not node.auxiliary]
            self.width = max(x) - min(x)
            self.height = max(y) - min(y)
            return False

        if self.debug & 4:
            print('Creating graphs')
        placer = schemplacer(self.elements, self.nodes, method, debug)
        if self.debug & 4:
            print('Solving graphs')
        self.width, self.height = placer.solve(self.node_spacing)

        return True

    def _write_coordinates(self):

        s = ''
        for n in self.nodes.values():
            s += r'  \coordinate (%s) at (%s);''\n' % (n.s, n.pos)
        return s

    def _draw_components(self, **kwargs):

        s = ''
        for elt in self.elements.values():
            if elt.ignore:
                continue
            if elt.directive:
                for key, val in elt.opts.items():
                    # Local opts overrides global opts in kwargs
                    kwargs[key] = val

            s += elt.draw(**kwargs)
        return s

    def _draw_nodes(self, **kwargs):

        s = ''
        for elt in self.elements.values():
            if elt.ignore:
                continue
            if elt.directive:
                for key, val in elt.opts.items():
                    # Local opts overrides global opts in kwargs
                    kwargs[key] = val

            s += elt.draw_nodes(**kwargs)
            s += elt.draw_pins()
        return s

    def _draw_node_labels(self, **kwargs):

        s = ''
        for elt in self.elements.values():
            if elt.ignore:
                continue
            if elt.directive:
                for key, val in elt.opts.items():
                    # Local opts overrides global opts in kwargs
                    kwargs[key] = val

            s += elt.draw_node_labels(**kwargs)
        return s

    def _draw_help_lines(self, help_lines, help_lines_color):

        start = Pos(-0.5, -0.5) * self.node_spacing
        stop = Pos(self.width + 0.5, self.height + 0.5) * self.node_spacing

        s = r'  \draw[help lines, %s] (%s) grid [xstep=%s, ystep=%s] (%s);''\n' % (
            help_lines_color, start, help_lines, help_lines, stop)

        return s

    def _tikz_draw(self, style_args='', **kwargs):

        if (self.debug & 64):
            import pdb
            pdb.set_trace()

        method = kwargs.pop('method', 'graph')
        preamble = kwargs.pop('preamble', None)
        postamble = kwargs.pop('postamble', None)
        options = kwargs.pop('options', None)
        help_lines = float(kwargs.pop('help_lines', 0))
        help_lines_color = kwargs.pop('help_lines_color', 'blue')

        self._positions_calculate(method, self.debug, **kwargs)

        # Note, scale does not scale the font size.
        opts = ['scale=%.2f' % self.scale,
                'transform shape',
                '/tikz/circuitikz/bipoles/length=%.2fcm' % self.cpt_size]
        opts.append(style_args)
        if options is not None:
            opts.append(options)

        global_options = ['font', 'voltage_dir', 'color']

        for opt in global_options:
            if opt in kwargs:
                opts.append(opt.replace('_', ' ') + '=' + kwargs.pop(opt))

        s = r'\begin{tikzpicture}[%s]''\n' % ', '.join(opts)

        # Add preamble
        if preamble:
            s += '  ' + preamble + '\n'

        if help_lines != 0:
            s += self._draw_help_lines(help_lines, help_lines_color)

        # Write coordinates.
        s += self._write_coordinates()

        # Draw components
        s += self._draw_components(**kwargs)

        # Draw nodes (do not do concurrently with _draw_components
        # since get a short piece of wire through ocirc nodes)
        s += self._draw_nodes(**kwargs)

        # Draw node labels
        s += self._draw_node_labels(**kwargs)

        # Add postamble
        if postamble:
            s += '  ' + postamble + '\n'

        s += r'\end{tikzpicture}''\n'

        return s

    def tikz_draw(self, filename, **kwargs):

        root, ext = path.splitext(filename)

        self.debug = kwargs.pop('debug', 0)
        style = kwargs.pop('style', 'american')
        self.dpi = float(kwargs.pop('dpi', PNG_DPI))
        self.cpt_size = float(kwargs.pop('cpt_size', 1.2))
        self.node_spacing = float(kwargs.pop('node_spacing', 2.0))
        self.scale = float(kwargs.pop('scale', 1.0))
        png_converter = kwargs.pop('png_converter', None)

        include = ''
        if 'include' in kwargs:
            include += kwargs.pop('include')
        if 'includefile' in kwargs:
            file = open(kwargs.pop('includefile'))
            include += file.read()
            file.close()

        if style == 'american':
            style_args = 'american currents, american voltages'
        elif style == 'british':
            style_args = 'american currents, european voltages'
        elif style == 'european':
            style_args = ('european currents, european voltages'
                          ', european inductors, european resistors')
        else:
            raise ValueError('Unknown style %s' % style)

        # For debugging when do not want to write to file
        nosave = kwargs.pop('nosave', False)

        latexrunner = LatexRunner((self.debug & 2) != 0)

        content = self._tikz_draw(style_args=style_args, **kwargs)

        if nosave:
            return

        if self.debug & 1:
            date, version = latexrunner.find_circuitikz_version()
            if date is None:
                raise RuntimeError('circuitikz is not installed')

            print('circuitikz version %s (%s)' % (version, date))
            print('width=%d, height=%d, dpi=%d, cpt_size=%.2f, node_spacing=%.2f, scale=%.2f'
                  % (self.width, self.height, self.dpi,
                     self.cpt_size, self.node_spacing, self.scale))
            print('Nodes:')
            for node in self.nodes.values():
                node.debug()

        if ext in ('.pytex', '.schtex', '.pgf'):
            if include != '':
                warn('Include option ignored for non-standalone file')
            open(filename, 'w').write(content)
            return

        # Need amsmath for operatorname
        template = ('\\documentclass[a4paper]{standalone}\n'
                    '\\usepackage{amsmath}\n'
                    '\\usepackage{circuitikz}\n'
                    '\\usetikzlibrary{fit, shapes, arrows, patterns, decorations.text, decorations.markings}\n'
                    '%s\n'
                    '\\begin{document}\n%s\\end{document}')
        content = template % (include, content)
        tex_filename = filename.replace(ext, '.tex')
        open(tex_filename, 'w').write(content)

        if ext == '.tex':
            return

        pdf_filename = tex_filename.replace('.tex', '.pdf')
        latexrunner.run(tex_filename)

        date, version = latexrunner.extract_circuitikz_version(tex_filename)
        if date is None:
            raise RuntimeError('circuitikz is not installed')

        if version < CIRCUITIKZ_MIN_VERSION:
            warn('Have circuitikz version %s; should upgrade to %s or later'
                 % (version, CIRCUITIKZ_MIN_VERSION))

        if not (self.debug & 1):
            latexrunner.cleanup(tex_filename, pdf_filename)

        if not path.exists(pdf_filename):
            raise RuntimeError('Could not generate %s with pdflatex' %
                               pdf_filename)

        if ext == '.pdf':
            return

        if ext == '.svg':
            pdfconverter = PDFConverter((self.debug & 2) != 0)
            pdfconverter.to_svg(pdf_filename, root + '.svg')
            if not (self.debug & 1):
                remove(pdf_filename)
            return

        if ext == '.png':
            pdfconverter = PDFConverter((self.debug & 2) != 0)
            pdfconverter.to_png(pdf_filename, root + '.png', dpi=self.dpi,
                                method=png_converter)
            if not (self.debug & 1):
                remove(pdf_filename)
            return

        raise RuntimeError('Cannot create file of type %s' % ext)

    def draw(self, filename=None, **kwargs):
        """
        filename specifies the name of the file to produce.  If None,
        the schematic is displayed on the screen.

        Note, if using Jupyter, then need to first issue command %matplotlib inline

        kwargs include:
           'label_ids': True to show component ids
           'label_values': True to display component values
           'annotate_values': True to display component values as separate label
           'draw_nodes': True to show all nodes,
             False or 'none' to show no nodes,
             'primary' to show primary nodes,
             'connections' to show nodes that connect more than two components,
             'all' to show all nodes
           'label_nodes': True to label all nodes,
             False or 'none' to label no nodes,
             'primary' to label primary nodes (nodes without an underscore),
             'alpha' to label nodes starting with a letter,
             'pins' to label nodes that are pins on a chip,
             'all' to label all nodes,
           'anchor': where to position node label (default south east)
           'include': name of file to include before \\begin{document}
           'style': 'american', 'british', or 'european'
           'scale': schematic scale factor, default 1.0
           'node_spacing': spacing between component nodes, default 2.0
           'cpt_size': size of a component, default 1.5
           'dpi': dots per inch for png files, default 300
           'help_lines': distance between lines in grid, default 0 (disabled)
           'debug': non-zero to display debug information
        """

        # None means don't care, so remove.  Use list so can remove
        # entries from kwargs
        for key in list(kwargs):
            if kwargs[key] is None:
                kwargs.pop(key)

        # Remove options that may be overridden by arguments to draw
        for elt in self.elements.values():
            for key in list(elt.opts):
                if key in kwargs:
                    elt.opts.remove(key)

        # Default options
        opts = SchematicOpts()
        for key, val in opts.items():
            if key not in kwargs:
                kwargs[key] = val

        # Global options (at end of netlist for historical reasons)
        # These overwrite the default options.  Skip blank lines.
        for elt_name in reversed(self.elements):
            elt = self.elements[elt_name]
            if not elt.directive:
                break
            if elt.string.strip().startswith(';;'):
                continue
            if elt.string.strip().startswith(';'):
                for key, val in elt.opts.items():
                    if key != 'thickness':
                        # val is a str
                        kwargs[key] = val
                break

        def in_ipynb():
            try:
                ip = get_ipython()
                cfg = ip.config

                kernapp = cfg['IPKernelApp']

                # Check if processing ipynb file for Jupyter notebook.
                if 'connection_file' in kernapp:
                    return True
                elif kernapp and kernapp['parent_appname'] == 'ipython-notebook':
                    return True
                else:
                    return False
            except (NameError, KeyError):
                return False

        if not self.hints:
            warn('No schematic drawing hints provided!')

        png = kwargs.pop('png', False)
        svg = kwargs.pop('svg', False)

        if in_ipynb() and filename is None:

            if not png and not svg:
                svg = False

            if svg:
                try:
                    from IPython.display import SVG, display_svg

                    svgfilename = tmpfilename('.svg')
                    self.tikz_draw(svgfilename, **kwargs)

                    # Create and display SVG image object.
                    # Note, there is a problem displaying multiple SVG
                    # files since the later ones inherit the namespace of
                    # the first ones.
                    display_svg(SVG(filename=svgfilename))
                    return
                except:
                    pass

            from IPython.display import Image, display_png

            png_filename = tmpfilename('.png')
            self.tikz_draw(png_filename, **kwargs)

            # Create and display PNG image object.
            # There are two problems:
            # 1. The image metadata (width, height) is ignored
            #    when the ipynb file is loaded.
            # 2. The image metadata (width, height) is not stored
            #    when the ipynb file is written non-interactively.
            width, height = png_image_size(png_filename)
            # width, height specify the image dimension in pixels
            display_png(Image(filename=png_filename,
                              width=width, height=height))
            if not (self.debug & 1):
                remove(png_filename)
            return

        if filename is None:
            filename = tmpfilename('.png')
            # Thicken up lines to reduce aliasing causing them to
            # disappear, especially when using pdftoppm.
            self.tikz_draw(filename=filename,
                           options='/tikz/circuitikz/bipoles/thickness=2',
                           **kwargs)
            display_matplotlib(filename, self.dpi)
            if not (self.debug & 1):
                remove(filename)
            return

        self.tikz_draw(filename=filename, **kwargs)

    def pdb(self):
        """Enter the python debugger."""

        import pdb
        pdb.set_trace()
        return self


def test():

    sch = Schematic()

    sch.add('P1 1 0.1')
    sch.add('R1 1 3; right')
    sch.add('L1 3 2; right')
    sch.add('C1 3 0; down')
    sch.add('P2 2 0.2')
    sch.add('W 0.1 0; right')
    sch.add('W 0 0.2; right')

    sch.draw()
    return sch



"""
This module performs schematic component placement using a graph algorithm.

Copyright 2021--2023 Michael Hayes, UCECE
"""

from .schemgraph import Graph
from .schemmisc import Pos
from .schemplacerbase import SchemPlacerBase

# Components are positioned using two graphs; one graph for
# the x direction and the other for the y direction.
#
# There is a naming confusion.  We have network nodes (electrical
# nodes) and nodes in the graph used for component placement.
# Let's call the latter gnodes.  The names of these gnodes are a
# tuple of the common electrical nodes.
#
# x and y component positioning are performed independently.  Let's
# consider the x or horizontal positioning.  There are three stages:
#   1. Component nodes that share a y position are linked; this can
#      occur, for example, for a vertically oriented component.
#      This helps to reduce the size of the graph.
#   2. The x positions of the components are used to determine the
#      graph edges.
#   3. The longest path through the graph is found and the x positions
#      of the nodes are assigned based on the distance along the
#      longest path.


class SchemGraphPlacer(SchemPlacerBase):

    def __init__(self, elements, nodes, debug=0):

        self.elements = elements
        self.nodes = nodes
        self.debug = debug

        self.xgraph = Graph('horizontal', nodes, debug)
        self.ygraph = Graph('vertical', nodes, debug)



"""This module provides support for nodes.

Copyright 2020--2023 Michael Hayes, UCECE

"""

from .immittancemixin import ImmittanceMixin


class DummyNode:

    def __init__(self, name):
        self.name = name


class Node(ImmittanceMixin):

    def __init__(self, cct, name):

        self.cct = cct
        self._name = name
        self.pos = None
        parts = name.split('_')
        self.rootname = parts[0] if name[0] != '_' else name
        self.primary = len(parts) == 1
        # List of elements connected to this node.
        self._connected = []
        self._count = 0

    def __repr__(self):
        return "Node('%s')" % self.name

    def __str__(self):

        if self.pos is not None:
            # Note, need to have float type otherwise 0 becomes empty string.
            x = str(round(self.x, 2)).rstrip('0').rstrip('.')
            y = str(round(self.y, 2)).rstrip('0').rstrip('.')

            return '%s@(%s, %s)' % (self.name, x, y)
        else:
            return self.name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        # Use new node name for nodes
        self.cct.nodes[name] = self.cct.nodes.pop(self._name)
        self._name = name

    @property
    def x(self):

        if self.pos is None:
            raise ValueError('Node position undefined')
        return self.pos.x

    @property
    def y(self):

        if self.pos is None:
            raise ValueError('Node position undefined')
        return self.pos.y

    @property
    def V(self):
        """Node voltage with respect to ground."""

        return self.cct.get_Vd(self.name, '0')

    @property
    def v(self):
        """Node time-domain voltage with respect to ground."""

        return self.cct.get_vd(self.name, '0')

    @property
    def admittance(self):
        """Driving-point admittance between node and ground."""

        return self.cct.admittance(self.name, '0')

    @property
    def impedance(self):
        """Driving-point impedance between node and ground."""

        return self.cct.impedance(self.name, '0')

    def append(self, cpt):

        if cpt.type not in ('A', ):
            self._count += 1

        self._connected.append(cpt)

    def remove(self, cpt):

        if len(self._connected) == 0:
            raise RuntimeError(
                'Removing node %s with no connections' % str(self))

        for c in self._connected:
            if c.name == cpt.name:
                self._connected.remove(c)
                break
        if cpt.type not in ('A', ):
            self._count -= 1

        if self.count == 0:
            self.cct.nodes._remove(self.name)

    @property
    def count(self):
        """Number of elements electrically connected to the node.  This
        includes wires but not annotations.

        """

        return self._count

    def oneport(self, node=0):
        """Create oneport object with respect to specified node
        (default ground)."""

        return self.cct.oneport(self.name, node)

    def thevenin(self, node=0):
        """Create Thevenin oneport object with respect to specified
        node (default ground)."""

        return self.cct.thevenin(self.name, node)

    def norton(self, node=0):
        """Create Norton oneport object with respect to specified node
        (default ground)."""

        return self.cct.norton(self.name, node)

    @property
    def connected(self):
        """Return list of components connected to the node."""

        return self._connected

    def is_connected(self, cpt):
        """Return True if cpt is connected to the node."""

        if isinstance(cpt, str):
            for cpt1 in self.connected:
                if cpt1.name == cpt:
                    return True
            return False

        return cpt in self.connected

    def is_wired_to(self, node):
        """Return True if the nodes are wired together."""

        node = str(node)
        for nodes in self.cct.equipotential_nodes.values():
            if self.name in nodes and node in nodes:
                return True
        return False

    def wired_to(self):
        """Return list of names of nodes that are wired to this node."""

        for nodes in self.cct.equipotential_nodes.values():
            if self.name in nodes:
                nodes.remove(self.name)
                return nodes
        return []

    @property
    def is_port(self):
        """Return True if node is a port"""

        for cpt in self.connected:
            if cpt.type == 'P':
                return True
        return False

    @property
    def port(self):

        return self.is_port

    @property
    def is_dangling(self):
        """Return True if node has a single electrical connection"""

        return self._count <= 1

    @property
    def is_ground(self):
        """Return True if node is a ground"""

        return self.name.startswith('0')

    def debug(self):

        s = str(self) + ', count=%s' % self.count

        names = [cpt.name for cpt in self.connected]

        s += ', cpts=[%s]' % ', '.join(names) + '\n'

        return s



"""This module provides susceptance support.

Copyright 2021--2022 Michael Hayes, UCECE

"""
from .expr import expr
from .sym import j
from .units import u as uu
from warnings import warn


def susceptance(arg, **assumptions):
    """Create an admittance class for the specified susceptance.

    The susceptance, B, is defined as the imaginary part of the admittance.

    Y(omega) = G(omega) + j * B(omega)"""

    expr1 = expr(arg, frequency=True, **assumptions)
    if expr1.is_laplace_domain:
        warn('Specifying Laplace domain for susceptance: %s' % expr1)

    if expr1.is_imaginary:
        warn('Susceptance %s should be real' % expr1)

    expr1.units = uu.siemens
    return expr1



"""This module provides the cached_property decorator.

Copyright 2021 Michael Hayes, UCECE
"""

try:
    # Requires python3.8
    from functools import cached_property
except:
    from property_cached import cached_property

from functools import lru_cache



"""This module provides continuous-time filter support.

Copyright 2022-2023 Michael Hayes, UCECE

"""

from .expr import expr, equation, ExprTuple
from .differentialequation import DifferentialEquation
from .functions import Derivative, Function, exp, cos
from .texpr import TimeDomainExpression
from .transfer import transfer
from .symbols import t, s, omega0, j, pi, f
from .utils import isiterable
import sympy as sym


class LTIFilter(object):

    def __init__(self, b, a):
        """Create continuous-time filter where `b` is a list or array of
        numerator coefficients and `a` is a list or array of
        denominator coefficients."""

        # For numerical stability it might be better to store filter
        # as ZPK form.

        if not isiterable(b):
            b = (b, )
        if not isiterable(a):
            a = (a, )

        self.a = ExprTuple(a)
        self.b = ExprTuple(b)

    def __repr__(self):
        """This is called by repr(expr).  It is used, e.g., when printing
        in the debugger."""

        return '%s(%s, %s)' % (self.__class__.__name__, self.b, self.a)

    @classmethod
    def from_ZPK(cls, Z, P, K=1):
        """Create LTIFilter given transfer function in zero-pole-gain form
        where `Z` is a list of zeroes, `P` is a list of poles, and `K`
        is a constant."""

        if not isiterable(Z):
            Z = (Z, )
        if not isiterable(P):
            P = (P, )

        N = expr(K)
        for z in Z:
            N *= (s - z)
        b = N.coeffs()

        D = expr(1)
        for p in P:
            D *= (s - p)
        a = D.coeffs()

        return cls(b, a)

    @classmethod
    def from_transfer_function(cls, H, normalize_a0=True):
        """Create LTIFilter given a transfer function."""

        H = transfer(H)
        if not H.is_rational_function:
            raise ValueError("Transfer function is not a rational function")

        N, D = H.as_N_D()

        bn = N.coeffs()
        an = D.coeffs()

        if normalize_a0:
            bn = [bx / an[0] for bx in bn]
            an = [ax / an[0] for ax in an]

        return cls(bn, an)

    def transfer_function(self):
        """Return continuous-time impulse response (transfer function) in
        s-domain."""

        from .sym import ssym

        a = self.a.sympy
        b = self.b.sympy

        Nl = len(a)
        Nr = len(b)

        # Numerator of H(s)
        num = sym.S.Zero
        for i in range(Nr):
            num += b[i] * ssym**(Nr - i - 1)

        # Denominator for H(s)
        denom = sym.S.Zero
        for k in range(Nl):
            denom += a[k] * ssym**(Nl - k - 1)

        # Collect with respect to positive powers of the variable s
        num = sym.collect(sym.expand(num), ssym)
        denom = sym.collect(sym.expand(denom), ssym)

        Hs = transfer(sym.simplify(num / denom))
        return Hs

    def impulse_response(self):
        """Return continuous-time impulse response (transfer function) in time
        domain."""

        H = self.transfer_function()
        return H(t)

    def step_response(self):
        """Return continuous-time step response in time domain."""

        H = self.transfer_function()
        G = H / s

        return G(t)

    def frequency_response(self, **assumptions):
        """Return frequency response."""

        return self.transfer_function().frequency_response(**assumptions)

    def angular_frequency_response(self, **assumptions):
        """Return angular frequency response."""

        return self.transfer_function().angular_frequency_response(**assumptions)

    def phase_response(self):

        H = self.frequency_response()
        return H.phase

    def group_delay(self):

        phi = self.phase_response()
        tau = -Derivative(phi, f) / (2 * pi)
        return tau.simplify().general()

    def phase_delay(self):

        phi = self.phase_response()
        tau = -phi / (2 * pi * f)
        return tau.simplify().general()

    def differential_equation(self, inputsym='x', outputsym='y'):
        """Return differential equation."""

        rhs = 0 * t
        lhs = 0 * t

        x = Function(inputsym)(t)
        y = Function(outputsym)(t)

        for m, bn in enumerate(reversed(self.b)):
            rhs += bn * Derivative(x, t, m)

        for m, an in enumerate(reversed(self.a)):
            lhs += an * Derivative(y, t, m)

        return DifferentialEquation(lhs, rhs, inputsym, outputsym)

    def sdomain_initial_response(self, ic=None):
        """Return s-domain response due to initial conditions.
           ic : list or tuple of initial conditions y(0), y'(0), y''(0), ...

        """

        from .sym import ssym

        if ic is None:
            ic = ()

        if not isiterable(ic):
            ic = (ic, )

        ic = ExprTuple(ic)

        Nl = len(self.a)
        if len(ic) != Nl - 1:
            raise ValueError('Expecting %d initial conditions, got %d'
                             % (Nl - 1, len(ic)))

        # Denominator for Yi(s)
        denom = self.a[0] * s**0
        num = 0 * s
        for k in range(1, Nl):
            aas = self.a[k] * s**(-k)
            denom += aas
            # Numerator for Yi(s)
            y0 = 0 * s
            for i in range(0, k):
                y0 += ic[i] * s**(i + 1)
            num += aas * y0

        # Collect with respect to positive powers of the variable s
        num = num.sympy * ssym**Nl
        denom = denom.sympy * ssym**Nl

        num = sym.collect(sym.expand(num), ssym)
        denom = sym.collect(sym.expand(denom), ssym)

        Ysi = expr(-sym.simplify(num / denom))
        Ysi.is_causal = True

        return Ysi

    def initial_response(self, ic=None):
        """Return response due to initial conditions in the time domain.
           ic : list of initial conditions y(0), y'(0), y''(0), ...
        """

        Ysi = self.sdomain_initial_response(ic)
        return Ysi(t)

    def response(self, x, ic=None, ni=None):
        """Calculate response of filter to input `x` given a list of initial conditions
        `ic` for time indexes specified by `ni`.  If `ni` is a tuple,
        this specifies the first and last (inclusive) time index.

        The initial conditions are valid prior to the time indices given by the ni
        `x` can be an expression, a sequence, or a list/array of values.
        """

        return 0

    def subs(self, *args, **kwargs):

        a = self.a.subs(*args, **kwargs)
        b = self.b.subs(*args, **kwargs)

        return self.__class__(b, a)

    def pdb(self):
        """Enter the python debugger."""

        import pdb
        pdb.set_trace()
        return self

    def discretize(self, method='bilinear', alpha=0.5):
        """Convert to a discrete-time linear time-invariant filter.

        The default method is 'bilinear'.  Other methods are:
        'impulse-invariance' 'bilinear', 'tustin', 'trapezoidal'
        'generalized-bilinear', 'gbf' controlled by the parameter
        `alpha`, 'euler', 'forward-diff', 'forward-euler'
        'backward-diff', 'backward-euler' 'simpson', 'matched-Z',
        'zero-pole-matching'"""

        H = self.transfer_function()

        return H.discretize(method, alpha).dlti_filter()

    @property
    def is_stable(self):
        """Return True if all the poles of the filters transfer function
        are within the unit circle.

        See also is_marginally_stable."""

        return self.transfer_function().is_stable

    @property
    def is_marginally_stable(self):
        """Return True if all the poles of the filters transfer function
        are within or on the unit circle.

        See also is_stable."""

        return self.transfer_function().is_marginally_stable

    @classmethod
    def bessel(cls, N=2, Wn=omega0, btype='lowpass'):
        """Create a Bessel filter of specified order `N`
        and angular cut-off frequency Wn."""

        from math import factorial

        # TODO, add band-pass and band-stop with list for Wn.
        omega0 = Wn

        denom = expr(0)
        for k in range(0, N + 1):

            a = factorial(2 * N - k) / (2 ** (N - k) *
                                        factorial(k) * factorial(N - k))
            denom += a * (s / omega0)**k

        numer = denom.limit(s, 0)

        if btype == 'lowpass':
            b = numer.coeffs()
            a = denom.coeffs()
        elif btype == 'highpass':
            H = (numer / denom).subs(s, omega0**2 / s)
            a = H.a
            b = H.b
        else:
            raise ValueError(
                'Bessel filter of btype %s not supported' % btype)

        return cls(b, a)

    @classmethod
    def butterworth(cls, N=2, Wn=omega0, btype='lowpass'):
        """Create a Butterworth filter of specified order `N`
        and angular cut-off frequency Wn."""

        # TODO, add band-pass and band-stop with list for Wn.
        omega0 = Wn

        numer = expr(1)
        denom = expr(1)
        for k in range(1, N + 1):
            sk = omega0 * exp(j * (2 * k + N - 1) * pi / (2 * N))
            numer *= omega0
            # Rewrite exp as cos to improve simplification.  For some reason
            # SymPy does not simplify exp(-j * pi / 4).
            sk = sk.rewrite(cos)
            denom *= (s - sk)

        # Can convert low-pass to high-pass with s = omega0**2 / s

        if btype == 'lowpass':
            b = numer.coeffs()
        elif btype == 'highpass':
            b = [0] * (N + 1)
            b[0] = 1
        else:
            raise ValueError(
                'Butterworth filter of btype %s not supported' % btype)

        a = denom.coeffs()

        return cls(b, a)


Bessel = LTIFilter.bessel
Butterworth = LTIFilter.butterworth
